# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import logging
import re

# Third Party
import pytest

# Local
from caikit.core.toolkit import wip_decorator
from caikit.core.toolkit.wip_decorator import (
    Action,
    TempDisableWIP,
    WipCategory,
    disable_wip,
    enable_wip,
    work_in_progress,
)

## Helpers #####################################################################


@pytest.fixture(autouse=True)
def with_wip():
    prev_val = wip_decorator._ENABLE_DECORATOR
    enable_wip()
    yield
    wip_decorator._ENABLE_DECORATOR = prev_val


msg_re = "<.*> is still in the {} phase and subject to change!"
expected_msg_beta = msg_re.format(WipCategory.BETA.name)
expected_msg_wip = msg_re.format(WipCategory.WIP.name)

## Tests #######################################################################

######################## Test decorator on function #########################################


def test_default_config_fn_no_arg(caplog):
    """Test decorator with no config on function with no arguments"""

    @work_in_progress
    def foo():
        pass

    with caplog.at_level(logging.WARN):
        foo()
        assert any(
            re.search(expected_msg_wip, record.message) for record in caplog.records
        )


def test_default_config_fn_arg(caplog):
    """Test decorator with no config on function with arguments"""

    @work_in_progress
    def foo(x):
        return x

    argument = 1
    with caplog.at_level(logging.WARN):
        result = foo(argument)
        assert any(
            re.search(expected_msg_wip, record.message) for record in caplog.records
        )
        assert argument == result


def test_default_config_fn_kwarg(caplog):
    """Test decorator with no config on function with argument and keyword argument"""

    @work_in_progress
    def foo(x, y=0, z=0):
        return x + y + z

    argument = 1
    kwargs = {"y": 1, "z": 1}
    expected_result = 3
    with caplog.at_level(logging.WARN):
        result = foo(argument, **kwargs)
        assert any(
            re.search(expected_msg_wip, record.message) for record in caplog.records
        )
        assert expected_result == result


def test_custom_config_fn_arg(caplog):
    """Test decorator with custom config and argument"""

    @work_in_progress(category=WipCategory.BETA)
    def foo(x):
        return x

    argument = 1
    with caplog.at_level(logging.WARN):
        result = foo(argument)
        assert any(
            re.search(expected_msg_beta, record.message) for record in caplog.records
        )
        assert argument == result


def test_temp_disable_wip_with_ctxt(caplog):
    """Test decorator with context based temp decorator disabling"""

    @work_in_progress
    def foo(y=0, z=0):
        return 1 + y + z

    with caplog.at_level(logging.WARN):
        # Check temp disable
        with TempDisableWIP():
            result = foo(1)
            assert result == 2
            # no log lines produced
            assert len(caplog.records) == 0


######################## Test decorator on class ############################################


def test_default_config_class_no_arg(caplog):
    """Test decorator with no config on class with no init args"""

    @work_in_progress
    class Foo:
        pass

    with caplog.at_level(logging.WARN):
        Foo()
        assert any(
            re.search(expected_msg_wip, record.message) for record in caplog.records
        )


def test_default_config_class_arg(caplog):
    """Test decorator with no config on class with init arguments"""

    @work_in_progress
    class Foo:
        def __init__(self, x) -> None:
            self.x = x

    argument = 1
    with caplog.at_level(logging.WARN):
        result = Foo(argument)
        assert any(
            re.search(expected_msg_wip, record.message) for record in caplog.records
        )
        assert argument == result.x


def test_default_config_class_kwarg(caplog):
    """Test decorator with no config on class with init argument and keyword argument"""

    @work_in_progress
    class Foo:
        def __init__(self, x, y=0, z=0) -> None:
            self.x = x + y + z

    argument = 1
    kwargs = {"y": 1, "z": 1}
    expected_result = 3
    with caplog.at_level(logging.WARN):
        result = Foo(argument, **kwargs)
        assert any(
            re.search(expected_msg_wip, record.message) for record in caplog.records
        )
        assert expected_result == result.x


def test_custom_config_class_arg(caplog):
    """Test decorator on class with custom config and init argument"""

    @work_in_progress(category=WipCategory.BETA)
    class Foo:
        def __init__(self, x) -> None:
            self.x = x

    argument = 1
    with caplog.at_level(logging.WARN):
        result = Foo(argument)
        assert any(
            re.search(expected_msg_beta, record.message) for record in caplog.records
        )
        assert argument == result.x


def test_inheritance_of_decorated_class(caplog):
    """Test decorator on a class that is inherited by another class
    still works.
    """

    @work_in_progress(category=WipCategory.BETA)
    class Foo:
        pass

    class Bar(Foo):
        def __init__(self, x) -> None:
            self.x = x

    argument = 1
    with caplog.at_level(logging.WARN):
        result = Bar(argument)
        assert any(
            re.search(expected_msg_beta, record.message) for record in caplog.records
        )
        assert argument == result.x


######################## Test separate actions #########################################


def test_error_action_raises_fn(caplog):
    """Test error action passed to decorator function raises"""
    func = lambda x: x
    work_in_progress(func, action=Action.ERROR)
    with pytest.raises(Exception):
        func()


def test_error_action_raises_class(caplog):
    """Test error action passed to decorator class raises"""

    @work_in_progress(action=Action.ERROR)
    class Foo:
        pass

    with pytest.raises(Exception):
        Foo()


######################## Test disable functionality #########################################


def test_decorator_can_be_disabled(caplog):
    """Test that decorator can be disabled using disable_wip function"""
    func = lambda x: x
    work_in_progress(func, action=Action.ERROR)
    # Sanity check that we get exception as expected
    with pytest.raises(Exception):
        func()
    # Disable decorator
    disable_wip()

    # Try again, this time there should not be any exception
    result = func(1)
    assert result == 1



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module holds the Task definitions for all common NLP tasks
"""

# Standard
from typing import Iterable

# Local
from ...core import TaskBase, task
from .data_model.classification import (
    ClassificationResults,
    ClassifiedGeneratedTextResult,
    ClassifiedGeneratedTextStreamResult,
    TokenClassificationResults,
    TokenClassificationStreamResult,
)
from .data_model.text import TokenizationResults, TokenizationStreamResult
from .data_model.text_generation import GeneratedTextResult, GeneratedTextStreamResult


@task(
    unary_parameters={"text": str},
    unary_output_type=GeneratedTextResult,
    streaming_output_type=Iterable[GeneratedTextStreamResult],
)
class TextGenerationTask(TaskBase):
    """The Text Generation Task is responsible for taking input prompting text
    and generating additional text from that prompt.
    """


@task(
    required_parameters={"text": str},
    output_type=ClassificationResults,
)
class TextClassificationTask(TaskBase):
    """The text classification task is responsible for assigning a label or class to text."""


@task(
    unary_parameters={"text": str},
    streaming_parameters={"text_stream": Iterable[str]},
    unary_output_type=TokenClassificationResults,
    streaming_output_type=Iterable[TokenClassificationStreamResult],
)
class TokenClassificationTask(TaskBase):
    """The token classification task is responsible for assigning a label to individual
    tokens in a document.
    """


@task(
    unary_parameters={"text": str},
    streaming_parameters={"text_stream": Iterable[str]},
    unary_output_type=TokenizationResults,
    streaming_output_type=Iterable[TokenizationStreamResult],
)
class TokenizationTask(TaskBase):
    """The tokenization task is responsible for splitting a document into tokens."""


@task(
    unary_parameters={"text": str},
    unary_output_type=ClassifiedGeneratedTextResult,
    streaming_output_type=Iterable[ClassifiedGeneratedTextStreamResult],
)
class ClassificationWithTextGenerationTask(TaskBase):
    """The classification with text generation task is responsible for taking
    input prompting text, generating additional text from that prompt and classifying
    the generated text based on detectors.
    """



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Custom to Google Docstring Converter

This script converts custom style docstrings to Google style. It extracts docstrings from a 
file or directory, converts them to Google style, and updates the file(s) with the converted 
docstrings.

Example Usage:
    python dsconverter.py my_script.py
    python dsconverter.py my_directory/
    tox -e dsconverter

Note:
    - The script assumes that the custom docstrings are written in triple quotes and follow a 
      specific structure.
    - It identifies different sections within the docstring such as Args, Returns, Notes, 
      Examples, Raises, and Attributes, and converts them accordingly.
    - It assumes docstrings sections are written in the following order 
      Args -> Returns -> Notes -> Examples -> Attributes -> Raises
    - The conversion includes formatting the descriptions and arguments, and reorganizing 
      the sections based on Google style.
    - Google Python Style Guide:
      http://google.github.io/styleguide/pyguide.html
"""


# Standard
import argparse
import ast
import os
import re
import textwrap


class CustomDocstringConverter:
    """A class for converting custom style docstrings to Google style."""

    @staticmethod
    def extract_docstrings_from_file(file_path):
        """Extracts docstrings from the specified file.
        Args:
            file_path (str): Path to the file.
        Returns:
            list: List of extracted docstrings.
        """
        with open(file_path, "r") as file:
            file_contents = file.read()

        # Parse the file contents into an abstract syntax tree (AST)
        tree = ast.parse(file_contents)

        # Collect all docstrings
        docstrings = []
        for node in ast.walk(tree):
            if (
                isinstance(node, (ast.FunctionDef, ast.ClassDef, ast.Module))
                and node.body
            ):
                # Check if the node has a docstring
                if isinstance(node.body[0], ast.Expr) and isinstance(
                    node.body[0].value, ast.Str
                ):
                    docstring = node.body[0].value.s
                    docstrings.append(docstring)

        return docstrings

    def convert_to_google_style(self, custom_docstring):
        """Converts a docstring to Google style.
        Args:
            custom_docstring (str): The original docstring.
        Returns:
            str: The converted docstring in Google style.
        """

        def ret_replacement(match):
            (white_space, data_type, desc_white_space, description) = match.groups()
            num_words_data_type = len(data_type.split())
            if num_words_data_type >= 10:
                return f"{white_space}{data_type}\n{desc_white_space}{description}"
            # If has already been converted, don't convert it again
            if ":" in data_type:
                return f"{white_space}{data_type}\n{desc_white_space}{description}"
            if description:
                # Clean up return and reformat it
                cleaned_description = re.sub(r"\s*\n\s*", " ", description.strip())
                if "\n" in white_space:
                    white_space = "\n" + (" " * (len(white_space) - 2))
                wrapped_lines = textwrap.wrap(
                    f"{white_space}{data_type}: {cleaned_description}\n",
                    width=80,
                    subsequent_indent=desc_white_space,
                )
                returns_white_space = " " * (len(white_space) - 5)

                # Accounts for the fact Returns: may be last section
                if "\n" in white_space:
                    if returns_last:
                        return (
                            "\n" + "\n".join(wrapped_lines) + "\n" + returns_white_space
                        )
                    return "\n" + "\n".join(wrapped_lines)

                if returns_last:
                    return "\n".join(wrapped_lines) + "\n" + returns_white_space

                return "\n".join(wrapped_lines)

            if returns_last:
                return f"{white_space}{data_type}\n{desc_white_space}"

            return f"{white_space}{data_type}\n"

        def arg_replacement(match):
            (
                white_space,
                name,
                arg_white_space,
                data_type,
                desc_white_space,
                description,
            ) = match.groups()
            num_words_data_type = len(data_type.split())
            num_ors_data_type = len(data_type.split("|")) - 1
            num_word_ors_data_type = len(data_type.split("or")) - 1
            num_commas_data_type = len(data_type.split(", ")) - 1
            num_arrows_data_type = len(data_type.split("->")) - 1
            num_words_data_type -= (
                (num_ors_data_type * 2)
                + (num_word_ors_data_type * 2)
                + (num_commas_data_type)
                + (num_arrows_data_type * 2)
            )
            # If re accidentally absorbs description as data type
            if num_words_data_type >= 2:
                if description:
                    return f"{white_space}{name}: {data_type}\n{desc_white_space}{description}"
                return f"{white_space}{name}: {data_type}\n{desc_white_space}"
            if description:
                # Safety check if description is input incorrectly
                if (len(white_space.strip("\n"))) == len(desc_white_space):
                    return f"{white_space}{name} ({data_type})\n{desc_white_space}{description}"
                # Clean up argument and reformat it
                cleaned_description = re.sub(r"\s*\n\s*", " ", description.strip())
                # Wrapped lines created an extra space, must remove it
                if "\n" in white_space:
                    white_space = "\n" + (" " * (len(white_space) - 2))
                wrapped_lines = textwrap.wrap(
                    f"{white_space}{name} ({data_type}): {cleaned_description}",
                    width=80,
                    subsequent_indent=desc_white_space,
                )
                # If there was a newline in white space, replace it
                if "\n" in white_space:
                    return "\n" + "\n".join(wrapped_lines)

                return "\n".join(wrapped_lines)

            return f"{white_space}{name} ({data_type})\n{desc_white_space}"

        converted_docstring = custom_docstring
        args = False
        returns = False
        notes = False
        examples = False
        raises = False
        attributes = False
        returns_last = False

        # Determine args, returns, notes start
        if "Args:" in custom_docstring:
            args_start = custom_docstring.find("Args:")
            args = True
        elif "args:" in custom_docstring:
            args_start = custom_docstring.find("args:")
            args = True

        if "Returns:" in custom_docstring:
            returns_start = custom_docstring.find("Returns:")
            returns = True
        elif "returns:" in custom_docstring:
            returns_start = custom_docstring.find("returns:")
            returns = True

        if "Notes:" in custom_docstring:
            notes_start = custom_docstring.find("Notes:")
            notes = True
        elif "notes:" in custom_docstring:
            notes_start = custom_docstring.find("notes:")
            notes = True

        if "Examples:" in custom_docstring:
            examples_start = custom_docstring.find("Examples:")
            examples = True
        elif "examples:" in custom_docstring:
            examples_start = custom_docstring.find("examples:")
            examples = True
        elif "Example Usage:" in custom_docstring:
            examples_start = custom_docstring.find("Example Usage:")
            examples = True

        if "Raises:" in custom_docstring:
            raises_start = custom_docstring.find("Raises:")
            raises = True

        if "Attributes:" in custom_docstring:
            attributes_start = custom_docstring.find("Attributes:")
            attributes = True

        # Find second triple quote (end of docstring)
        docstring_end = custom_docstring.find('"""', custom_docstring.find('"""') + 1)

        # Determine args ending
        if returns:
            args_end = returns_start
        elif notes:
            args_end = notes_start
        elif examples:
            args_end = examples_start
        elif attributes:
            args_end = attributes_start
        elif raises:
            args_end = raises_start
        else:
            args_end = docstring_end

        # Determine returns ending
        if notes:
            returns_end = notes_start
        elif examples:
            returns_end = examples_start
        elif attributes:
            returns_end = attributes_start
        elif raises:
            returns_end = raises_start
        else:
            returns_end = docstring_end
            returns_last = True

        # Extract the preceding white space, name, colon, data type, and following white
        # space of arguments using expressions
        arg_pattern = (
            r"(\s*)(\*\*\w+|\*\w+|\w+)(:\s*)(.*)\s*\n(\s*)((?!.*:)(?:.*(?:\n\5|$).*)*)?"
        )

        # Extract the data type and description of returns using regular expressions
        ret_pattern = r"(\s*)(.+)\n(\s+)((?:.*(?:\n(?!\n)|$).*)*)"

        if returns:
            returns_string = custom_docstring[returns_start + 8 : returns_end]

            # Replace the return pattern matches with the converted parts
            returns_string = re.sub(
                ret_pattern, ret_replacement, returns_string, flags=re.MULTILINE
            )
            converted_docstring = (
                custom_docstring[: returns_start + 8]
                + returns_string
                + custom_docstring[returns_end:]
            )

        if args:
            args_string = custom_docstring[args_start + 5 : args_end]

            # Replace args pattern matches with converted parts
            args_string = re.sub(
                arg_pattern, arg_replacement, args_string, flags=re.MULTILINE
            )
            converted_docstring = (
                converted_docstring[: args_start + 5]
                + args_string
                + converted_docstring[args_end:]
            )

        return converted_docstring

    def convert_docstrings(docstrings, file_path):
        """Converts docstrings to Google style and updates the file.
        Args:
            docstrings (list): List of docstrings to convert.
            file_path (str): Path to the file.
        """
        with open(file_path, "r") as file:
            file_contents = file.read()

        # Helper function to find the start and end position of a docstring
        def find_docstring_position(file_contents, docstring):
            start = file_contents.find(docstring)
            end = start + len(docstring)
            return start, end

        # Iterate over the docstrings and replace them in the file contents
        for docstring in docstrings:
            start, end = find_docstring_position(file_contents, docstring)
            converted_docstring = CustomDocstringConverter().convert_to_google_style(
                docstring
            )
            file_contents = (
                file_contents[:start] + converted_docstring + file_contents[end:]
            )

        # Update the file with the modified contents
        with open(file_path, "w") as file:
            file.write(file_contents)


def main():
    """The main function for running the Custom Docstring Converter."""

    # Create an ArgumentParser object
    parser = argparse.ArgumentParser(description="Custom Docstring Converter")

    # Add an argument for the file path
    parser.add_argument("file_path", type=str, help="Path to the file to convert")

    # Parse the command-line arguments
    args = parser.parse_args()

    # Extract the file path or directory path from the arguments
    path = args.file_path

    # Check if the path is a file or directory
    if os.path.isfile(path):
        # Path is a file, convert only that file
        print("Updated: ", path, "\n")
        docstrings = CustomDocstringConverter.extract_docstrings_from_file(path)
        CustomDocstringConverter.convert_docstrings(docstrings, path)
    elif os.path.isdir(path):
        counter = 0
        # Path is a directory, convert all .py files in the directory
        for root, dirs, files in os.walk(path):
            for file in files:
                if file.endswith(".py"):
                    file_path = os.path.join(root, file)
                    counter += 1
                    print("Updated: ", file_path)
                    docstrings = CustomDocstringConverter.extract_docstrings_from_file(
                        file_path
                    )
                    CustomDocstringConverter.convert_docstrings(docstrings, file_path)
        print("Done! Updated ", counter, " .py files")
    else:
        print("Invalid path provided.")


if __name__ == "__main__":
    main()



# Local
from .composite_module import CompositeModule
from .inner_module import InnerModule
from .list_implementation import ListModule
from .primitive_party_implementation import SamplePrimitiveModule
from .sample_implementation import SampleModule
from .streaming_module import StreamingModule



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from glob import glob
from typing import Any, List, Type, Union
import os
import sys

# Third Party
import grpc

# First Party
from py_to_proto.dataclass_to_proto import Annotated, OneofField
import alog

# Local
from caikit.core.data_model.base import DataBase
from caikit.core.data_model.dataobject import _make_oneof_init, make_dataobject
from caikit.core.data_model.streams.data_stream import DataStream
from caikit.core.exceptions import error_handler
from caikit.interfaces.common.data_model.stream_sources import (
    Directory,
    File,
    ListOfFiles,
    S3Files,
)
from caikit.runtime.service_generation.proto_package import get_runtime_service_package
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
import caikit

# import common explicitly since this module needs it
import caikit.interfaces.common

# This global holds the mapping of element types to their respective
# DataStreamSource wrappers so that the same message is not recreated
# unnecessarily
_DATA_STREAM_SOURCE_TYPES = {}

log = alog.use_channel("DSTRM-SRC")
error = error_handler.get(log)


class DataStreamSourceBase(DataStream):
    """This base class acts as a sentinel so that dynamically generated data
    stream source classes can be identified programmatically.
    """

    def __init__(self):
        super().__init__(self._generator)

    def _generator(self):
        stream = self.to_data_stream()
        return stream.generator_func(*stream.generator_args, **stream.generator_kwargs)

    def __getstate__(self) -> bytes:
        """A DataStreamSource is pickled by serializing its source
        representation. This is particularly useful when sharing data streams
        across subprocesses to run training in an isolated process.
        """
        return self.to_binary_buffer()

    def __setstate__(self, pickle_bytes: bytes):
        """Unpickling a DataStreamSource basically involves unpacking the
        serialized source representation. The catch is that the oneof is
        represented strangely in __dict__, so we need to explicitly set all
        oneof members.
        """
        new_inst = self.__class__.from_binary_buffer(pickle_bytes)
        setattr(self, new_inst.which_oneof("data_stream"), new_inst.data_stream)
        self.generator_func = self._generator
        self.generator_args = tuple()
        self.generator_kwargs = {}

    # pylint: disable=too-many-return-statements
    def to_data_stream(self) -> DataStream:
        """Convert to the target data stream type based on the source type"""

        # Determine which of the value types is set
        set_field = None
        for field_name in self.get_proto_class().DESCRIPTOR.fields_by_name:
            if getattr(self, field_name) is not None:
                assert (
                    set_field is None
                ), "Found DataStreamSource with multiple sources set"
                set_field = field_name

        # If no field is set, return an empty DataStream
        if set_field is None:
            log.debug3("Returning empty data stream")
            return DataStream.from_iterable([])

        # If a S3 pointer is given, raise not implemented
        if set_field == "s3files":
            error(
                "<COR80419785E>",
                NotImplementedError(
                    "S3Files are not implemented as stream sources in this runtime."
                ),
            )

        # If jsondata, pull from the data elements directly
        if set_field == "jsondata":
            log.debug3("Pulling data stream from inline json")
            return DataStream.from_iterable(self.jsondata.data)

        # If jsonfile, attempt to read the file and pull in the data from there
        if set_field == "file":
            return self._create_data_stream_from_file(fname=self.file.filename)

        # If list of files, attempt to read all files and combine datastreams from all
        if set_field == "listoffiles":
            # combined list of data streams that we will chain and send back
            data_stream_list = []
            for fname in self.listoffiles.files:
                data_stream_list.append(self._create_data_stream_from_file(fname=fname))

            return DataStream.chain(data_stream_list).flatten()

        # If directory, attempt to read an element from each file in the dir
        if set_field == "directory":
            dirname = self.directory.dirname
            full_dirname = self._get_resolved_source_path(dirname)
            extension = self.directory.extension or "json"
            if not dirname or not os.path.isdir(full_dirname):
                raise CaikitRuntimeException(
                    grpc.StatusCode.INVALID_ARGUMENT,
                    f"Invalid {extension} directory source file: {full_dirname}",
                )
            files_with_ext = list(glob(os.path.join(full_dirname, "*." + extension)))
            # make sure at least 1 file with the given extension exists
            if len(files_with_ext) == 0:
                raise CaikitRuntimeException(
                    grpc.StatusCode.INVALID_ARGUMENT,
                    f"directory {dirname} contains no source files with extension {extension}",
                )
            if extension == "json":
                return DataStream.from_json_collection(full_dirname, extension).map(
                    self._to_element_type
                )
            if extension == "csv":
                return DataStream.from_csv_collection(full_dirname).map(
                    self._to_element_type
                )
            if extension == "jsonl":
                return DataStream.from_jsonl_collection(full_dirname).map(
                    self._to_element_type
                )
            raise CaikitRuntimeException(
                grpc.StatusCode.INVALID_ARGUMENT,
                f"Extension not supported! {extension}",
            )

    @classmethod
    def _create_data_stream_from_file(cls, fname: str) -> DataStream:
        """Create a data stream object by deducing file extension
        and reading the file accordingly"""

        _, extension = os.path.splitext(fname)
        if not extension:
            return cls._load_from_file_without_extension(fname)

        full_fname = cls._get_resolved_source_path(fname)
        log.debug3("Pulling data stream from %s file [%s]", extension, full_fname)

        if not fname or not os.path.isfile(full_fname):
            raise CaikitRuntimeException(
                grpc.StatusCode.INVALID_ARGUMENT,
                f"Invalid {extension} data source file: {fname}",
            )
        if extension == ".json":
            stream = DataStream.from_json_array(full_fname).map(cls._to_element_type)
            # Iterate once to make sure this is a json array
            stream.peek()
            return stream
        if extension == ".csv":
            return DataStream.from_header_csv(full_fname).map(cls._to_element_type)
        if extension == ".jsonl":
            return DataStream.from_jsonl(full_fname).map(cls._to_element_type)
        raise CaikitRuntimeException(
            grpc.StatusCode.INVALID_ARGUMENT,
            f"Extension not supported! {extension}",
        )

    @classmethod
    def _load_from_file_without_extension(cls, fname) -> DataStream:
        """Similar to _create_data_stream_from_file, but we don't have a file extension to work
        with. Attempt to create a data stream using one of a few well-known formats"""
        full_fname = cls._get_resolved_source_path(fname)
        log.debug3("Attempting to guess file type for file: %s", full_fname)
        for factory_method in (
            DataStream.from_json_array,
            DataStream.from_jsonl,
            DataStream.from_header_csv,
            DataStream.from_multipart_file,
        ):
            try:
                stream = factory_method(full_fname).map(cls._to_element_type)
                # Iterate once and assume we have the correct file type if this works
                stream.peek()
                return stream
            except Exception as e:  # pylint: disable=broad-exception-caught
                # Catch any exception: it's hard to know which all could be thrown by any of the
                # formatters
                log.debug3(
                    "Failed to load file %s using data stream factory method %s: %s",
                    full_fname,
                    factory_method,
                    e,
                    exc_info=True,
                )
        raise CaikitRuntimeException(
            grpc.StatusCode.INVALID_ARGUMENT,
            f"Could not load input file with no extension: {full_fname}",
        )

    @classmethod
    def _to_element_type(cls, raw_element: Any) -> "ElementType":
        """Stream adapter to adapt from the raw json object data representations
        to the underlying data objects
        """
        if issubclass(cls.ELEMENT_TYPE, DataBase):
            return cls.ELEMENT_TYPE.from_json(raw_element)
        return raw_element

    @staticmethod
    def _get_resolved_source_path(input_path: str) -> str:
        """Get a fully resolved path, including any shared prefix"""
        # Get any configured prefix
        source_pfx = caikit.get_config().data_streams.file_source_base
        # If a prefix is configured, use it, otherwise return the path as is
        # NOTE: os.path.join will ignore the prefix if input_path is absolute
        return os.path.join(source_pfx, input_path) if source_pfx else input_path


def make_data_stream_source(data_element_type: Type) -> Type[DataBase]:
    """Dynamically create a data stream source message type that supports
    pulling an iterable of the given type from all valid data stream sources
    """
    log.debug2("Looking for DataStreamSource[%s]", data_element_type)
    if data_element_type not in _DATA_STREAM_SOURCE_TYPES:
        cls_name = _make_data_stream_source_type_name(data_element_type)
        log.debug("Creating DataStreamSource[%s] -> %s", data_element_type, cls_name)

        # Set up the "sub class." In python, this is the same as creating a
        # standalone class with a __qualname__ that nests it under a parent
        # class. We do this outside the declaration of the parent class so that
        # this class can be referenced within the Union annotation for the
        # outer class itself. This class needs to be created dynamically because
        # it encapsulates type information about the elements of the data stream.
        package = get_runtime_service_package()
        JsonData = make_dataobject(
            package=package,
            proto_name=f"{cls_name}JsonData",
            name="JsonData",
            attrs={"__qualname__": f"{cls_name}.JsonData"},
            annotations={"data": List[data_element_type]},
        )

        # Create the outer class that encapsulates the Union (oneof) or the
        # various types of input sources
        data_object = make_dataobject(
            package=package,
            name=cls_name,
            bases=(DataStreamSourceBase,),
            attrs={
                "ELEMENT_TYPE": data_element_type,
                JsonData.__name__: JsonData,
                File.__name__: File,
                ListOfFiles.__name__: ListOfFiles,
                Directory.__name__: Directory,
                S3Files.__name__: S3Files,
            },
            annotations={
                "data_stream": Union[
                    Annotated[JsonData, OneofField(JsonData.__name__.lower())],
                    Annotated[File, OneofField(File.__name__.lower())],
                    Annotated[ListOfFiles, OneofField(ListOfFiles.__name__.lower())],
                    Annotated[Directory, OneofField(Directory.__name__.lower())],
                    Annotated[S3Files, OneofField(S3Files.__name__.lower())],
                ],
            },
        )

        # Add this data stream source to the common data model and the module
        # where it was declared
        setattr(
            caikit.interfaces.common.data_model,
            cls_name,
            data_object,
        )
        setattr(
            sys.modules[data_object.__module__],
            cls_name,
            data_object,
        )

        # Add an init that sequences the initialization so that
        # DataStreamSourceBase is initialized after DataBase
        orig_init = _make_oneof_init(data_object)

        def __init__(self, *args, **kwargs):
            try:
                orig_init(self, *args, **kwargs)
            except TypeError as err:
                raise CaikitRuntimeException(
                    grpc.StatusCode.INVALID_ARGUMENT, str(err)
                ) from err
            DataStreamSourceBase.__init__(self)

        setattr(data_object, "__init__", __init__)

        _DATA_STREAM_SOURCE_TYPES[data_element_type] = data_object

    # Return the global stream source object for this element type
    return _DATA_STREAM_SOURCE_TYPES[data_element_type]


def _make_data_stream_source_type_name(data_element_type: Type) -> str:
    """Make the name for data stream source class that wraps the given type"""
    element_name = data_element_type.__name__
    return "DataStreamSource{}".format(element_name[0].upper() + element_name[1:])



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The LocalModelInitializer loads a model locally, optionally with a non-local backend
if the given module provides multiple backend-specific implementations.

Configuration for LocalModelInitializer lives under the config as follows:

model_management:
    initializers:
        <initializer name>:
            type: LOCAL
            config:
                # List of module backend configurations in priority order
                backend_priority:
                    - type: LOCAL
                      config: {}
"""
# Standard
from typing import Callable, List, Optional
import copy
import inspect

# First Party
import aconfig
import alog

# Local
from ..exceptions import error_handler
from ..module_backends import BackendBase, backend_types
from ..modules import ModuleBase, ModuleConfig
from ..modules.decorator import SUPPORTED_LOAD_BACKENDS_VAR_NAME
from ..registries import (
    module_backend_classes,
    module_backend_registry,
    module_backend_types,
)
from .model_initializer_base import ModelInitializerBase

log = alog.use_channel("LLOAD")
error = error_handler.get(log)


class LocalModelInitializer(ModelInitializerBase):
    __doc__ = __doc__

    name = "LOCAL"

    def __init__(self, config: aconfig.Config, instance_name: str):
        """Construct with the config"""
        self._instance_name = instance_name
        self._backends = []
        backend_priority = config.get(
            "backend_priority",
            [aconfig.Config({"type": "LOCAL"}, override_env_vars=False)],
        )
        for i, backend_config in enumerate(backend_priority):
            error.value_check(
                "<COR48633635E>",
                "type" in backend_config,
                "All backend priority configs must have a 'type' field",
            )
            backend_type = backend_config.type
            error.value_check(
                "<COR72281596E>",
                backend_type in module_backend_types(),
                "Invalid backend [{}] found at backend_priority index [{}]",
                backend_type,
                i,
            )

            log.debug("Configuring backend (%d)[%s]", i, backend_type)
            backend_instance_config = backend_config.get("config", {})
            log.debug3(
                "Backend (%d)[%s] config: %s", i, backend_type, backend_instance_config
            )

            backend_class = module_backend_classes().get(backend_type)
            error.value_check(
                "<COR64618509E>",
                len(self._backends) == i,
                "({})[{}] already configured",
                i,
                backend_type,
            )
            error.value_check(
                "<COR39517372E>",
                backend_class is not None,
                "Unsupported backend type {}",
                backend_type,
            )
            if not isinstance(backend_class, type) and issubclass(
                backend_class, BackendBase
            ):
                error(
                    "<COR05184600E>",
                    TypeError(
                        f"Backend {backend_class} is not derived from BackendBase"
                    ),
                )

            log.debug2("Performing config for (%d)[%s]", i, backend_type)
            backend_instance = backend_class(backend_instance_config)

            # Add configuration to backends as per individual module requirements
            self._configure_backend_overrides(backend_type, backend_instance)

            # Add the instance to the registry
            self._backends.append(backend_instance)

        log.debug2(
            "All configured backends for %s: %s", self._instance_name, self._backends
        )

    def init(self, model_config: ModuleConfig, **kwargs) -> Optional[ModuleBase]:
        """Given a ModelConfig, attempt to initialize it locally, possibly using
        a non-local backend

        Args:
            model_config (ModuleConfig): The in-memory model config object for
                the model to be initialized

        Returns:
            model (Optional[ModuleBase]): The in-memory ModuleBase instance that
                is ready to run
        """
        module_id = model_config.module_id
        model_path = model_config.model_path
        module_implementations = module_backend_registry().get(module_id, {})
        log.debug2(
            "Number of available backend implementations for %s found: %d",
            module_id,
            len(module_implementations),
        )
        # Look up the backend that this model was created with
        model_creation_backend = model_config.get("model_backend", backend_types.LOCAL)
        log.debug2("Model creation backend: %s", model_creation_backend)

        # Iterate through each backend in priority order and see if this module
        # can load with it
        loaded_model = None
        for load_backend in self._backends:

            # Look in the module's implementations for this backend type
            backend_impl_obj = module_implementations.get(load_backend.backend_type)
            if backend_impl_obj is None:
                log.debug3(
                    "Module %s does not support loading with %s",
                    module_id,
                    load_backend.backend_type,
                )
                continue

            # Grab the concrete module class for this backend and check to
            # see if this model's artifacts were created with a version of
            # the module that can be loaded with this backend.
            module_backend_impl = backend_impl_obj.impl_class
            supported_load_backends = self._get_supported_load_backends(
                module_backend_impl
            )
            if model_creation_backend in supported_load_backends:
                log.debug3(
                    "Attempting to load %s (module_id %s) with backend %s and class %s",
                    model_path,
                    module_id,
                    load_backend.backend_type,
                    module_backend_impl.__name__,
                )
                extra_kwargs = {}
                if self._supports_arg(module_backend_impl.load, "load_backend"):
                    extra_kwargs["load_backend"] = load_backend
                # Try loading with the ModuleConfig directly and fall back to
                # loading with the model_path for compatibility
                try:
                    loaded_model = module_backend_impl.load(
                        model_config,
                        **extra_kwargs,
                        **kwargs,
                    )
                except Exception as err:  # pylint: disable=broad-exception-caught
                    log.warning(
                        "<COR98539580W>",
                        "DEPRECATION: Loading %s failed with ModuleConfig. Using model_path. %s",
                        module_backend_impl.MODULE_ID,
                        err,
                    )
                    loaded_model = module_backend_impl.load(
                        model_path,
                        **extra_kwargs,
                        **kwargs,
                    )

                error.type_check("<COR40080753E>", ModuleBase, model=loaded_model)
                if loaded_model is not None:
                    log.debug2(
                        "Successfully loaded %s with backend %s",
                        model_path,
                        load_backend.backend_type,
                    )
                    loaded_model.set_load_backend(load_backend)
                    break

        # Return the loaded model if it was able to load
        return loaded_model

    @property
    def backends(self) -> List[BackendBase]:
        """Return an immutable view of the priority sequence of backends"""
        return copy.copy(self._backends)

    ## Implementation Details ##################################################

    @staticmethod
    def _supports_arg(load_fn: Callable, arg_name: str) -> bool:
        """A load function supports the load_backend kwarg IFF it has an arg
        explicitly named load_backend or it has a ** kwarg capture
        """
        sig = inspect.signature(load_fn)
        return arg_name in sig.parameters

    def _get_supported_load_backends(self, backend_impl: ModuleBase):
        """Function to get a list of supported load backends
        that the module supports

        Args:
            backend_impl: caikit.core.ModuleBase
                Module implementing the backend
        Returns:
            list(backend_types)
                list of backends that are supported for model load
        """

        # Get list of backends that are supported for load
        # NOTE: since code in a module can change anytime, its support
        # for various backend might also change, in which case,
        # it would be better to keep the backend information in the model itself
        # If module_backend is None, then we will assume that this model is not loadable in
        # any other backend
        return getattr(backend_impl, SUPPORTED_LOAD_BACKENDS_VAR_NAME, [])

    @staticmethod
    def _configure_backend_overrides(backend: str, backend_instance: object):
        """Function to go over all the modules registered in the MODULE_BACKEND_REGISTRY
        for a particular backend and configure their backend overrides

        Args:
            backend: str
                Name of the backend to select from registry
            backend_instance: object
                Initialized backend instance. This object should
                implement the `register_config` function which will be
                used to merge / iteratively configure the backend
        """
        # Go through all the modules registered with particular backend
        for module_id, module_type_mapping in module_backend_registry().items():
            if backend in module_type_mapping:
                # check if it contains any special config
                config = module_type_mapping[backend].backend_config_override
                error.type_check("<COR61136899E>", dict, config=config)
                if len(config) != 0:
                    # TODO: Add a check here to see if the backend has already started
                    backend_instance.register_config(config)
                else:
                    log.debug2(
                        "No backend overrides configured for %s module and %s backend",
                        module_id,
                        backend,
                    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logging top-level configuration for `caikit.core` library.
"""

# First Party
import alog

# Local
from caikit.config import get_config


def configure():
    """Utility function to initialize logging stack components.

    This uses the caikit configuration to determine how logging should be configured.

    This should _only_ be called in the context of executing a __main__ application.
    Configuring the logger from a library context may override log config that has
    already been set by the consuming application.
    """
    caikit_config = get_config()

    # For pretty format, build the formatter with the configured channel width
    if caikit_config.log.formatter == "pretty":
        formatter = alog.AlogPrettyFormatter(caikit_config.log.channel_width)
    else:
        # Otherwise just use the config string
        formatter = caikit_config.log.formatter

    # Initialize the python alog stack
    alog.configure(
        caikit_config.log.level,
        caikit_config.log.filters,
        formatter,
        caikit_config.log.thread_id,
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from inspect import isclass
from typing import Callable, Dict, Iterable, List, Type, TypeVar, Union
import collections
import dataclasses
import typing

# First Party
from alog import alog

# Local
from .data_model import DataStream
from .data_model.base import DataBase
from .exceptions import error_handler
from .signature_parsing import CaikitMethodSignature

log = alog.use_channel("TASK_BASE")
error = error_handler.get(log)

ProtoableInputTypes = Type[Union[int, float, str, bytes, bool, DataBase]]
ValidInputTypes = Union[
    ProtoableInputTypes, List[ProtoableInputTypes], DataStream[ProtoableInputTypes]
]

_InferenceMethodBaseT = TypeVar("_InferenceMethodBaseT", bound=Callable)

_STREAM_OUT_ANNOTATION = "__streaming_output_type"
_STREAM_PARAMS_ANNOTATION = "__streaming_params"
_UNARY_OUT_ANNOTATION = "__unary_output_type"
_UNARY_PARAMS_ANNOTATION = "__unary_params"


class TaskBase:
    """The TaskBase defines the interface for an abstract AI task

    An AI task is a logical function signature which, when implemented, performs
    a task in some AI domain. The key property of a task is that the set of
    required input argument types and the output value type are consistent
    across all implementations of the task.
    """

    @dataclasses.dataclass
    class InferenceMethodPtr:
        """Little container class that holds a method name and its flavor of streaming.
        i.e. the args to a `@TaskClass.taskmethod` decoration.
        """

        method_name: str  # the simple name of a method, like "run"
        input_streaming: bool
        output_streaming: bool

    deferred_method_decorators: Dict[
        Type["TaskBase"], Dict[str, List["TaskBase.InferenceMethodPtr"]]
    ] = {}

    @classmethod
    def taskmethod(
        cls, input_streaming: bool = False, output_streaming: bool = False
    ) -> Callable[[_InferenceMethodBaseT], _InferenceMethodBaseT]:
        """Decorates a module instancemethod and indicates whether the inputs and outputs should
        be handled as streams. This will trigger validation that the signature of this method
        is compatible with the task's definition of input and output types.

        The actual handling of validating the method and registering it is deferred until after
        the module class is created, which happens outside the context of this decoration.
        """

        def decorator(inference_method: _InferenceMethodBaseT) -> _InferenceMethodBaseT:
            cls.deferred_method_decorators.setdefault(cls, {})
            fq_mod_name = ".".join(
                [
                    inference_method.__module__,
                    *inference_method.__qualname__.split(".")[0:-1],
                ]
            )
            cls.deferred_method_decorators[cls].setdefault(fq_mod_name, [])
            cls.deferred_method_decorators[cls][fq_mod_name].append(
                TaskBase.InferenceMethodPtr(
                    method_name=inference_method.__name__,
                    input_streaming=input_streaming,
                    output_streaming=output_streaming,
                )
            )
            return inference_method

        return decorator

    @classmethod
    def deferred_method_decoration(cls, module: Type):
        """Runs the actual decoration logic that `taskmethod` would have run if the module class
        existed during its lifetime.

        Validates that all decorated methods match the task's API expectations, and stores the
        signatures on the module class for access later.
        """
        if cls.has_inference_method_decorators(module):
            keyname = _make_keyname_for_module(module)
            deferred_decorations = cls.deferred_method_decorators[cls][keyname]
            for decoration in deferred_decorations:
                signature = CaikitMethodSignature(module, decoration.method_name)
                cls.validate_run_signature(
                    signature, decoration.input_streaming, decoration.output_streaming
                )

                module._INFERENCE_SIGNATURES.append(
                    (decoration.input_streaming, decoration.output_streaming, signature)
                )

    @classmethod
    def has_inference_method_decorators(cls, module_class: Type) -> bool:
        """Utility that returns true iff a module has any `@TaskClass.taskmethod` decorations"""
        if cls not in cls.deferred_method_decorators:
            return False
        return (
            _make_keyname_for_module(module_class)
            in cls.deferred_method_decorators[cls]
        )

    @classmethod
    def validate_run_signature(
        cls,
        signature: CaikitMethodSignature,
        input_streaming: bool,
        output_streaming: bool,
    ) -> None:
        """Validates that the provided method signature meets the api constraints defined in this
        task, for the given streaming flavors.

        Raises:
            ValueError if no type annotations were provided on the method
            TypeError if the type annotations do not meet the task's api constraints
        """
        if not signature.parameters:
            raise ValueError(
                "Task could not be validated, no .run parameters were provided"
            )
        if signature.return_type is None:
            raise ValueError(
                "Task could not be validated, no .run return type was provided"
            )

        missing_required_params = [
            parameter_name
            for parameter_name in cls.get_required_parameters(input_streaming)
            if parameter_name not in signature.parameters
        ]
        if missing_required_params:
            raise TypeError(
                f"Required parameters {missing_required_params} not in signature for module: "
                f"{signature.module}"
            )

        type_mismatch_errors = []
        for parameter_name, parameter_type in cls.get_required_parameters(
            input_streaming
        ).items():
            signature_type = signature.parameters[parameter_name]
            if parameter_type != signature_type:
                if typing.get_origin(
                    signature_type
                ) == typing.Union and parameter_type in typing.get_args(signature_type):
                    continue
                if input_streaming and cls._is_iterable_type(parameter_type):
                    streaming_type = typing.get_args(parameter_type)[0]

                    for iterable_type in typing.get_args(signature_type):
                        if not cls._subclass_check(iterable_type, streaming_type):
                            raise TypeError(
                                f"Wrong input type for {parameter_name}, expected {parameter_type} \
                                  but got {signature_type}"
                            )
                else:
                    type_mismatch_errors.append(
                        f"Parameter {parameter_name} has type {signature_type} but type \
                            {parameter_type} is required"
                    )
        if type_mismatch_errors:
            raise TypeError(
                f"Wrong types provided for parameters to {signature.module}: {type_mismatch_errors}"
            )

        cls._raise_on_wrong_output_type(
            signature.return_type, signature.module, output_streaming
        )

    @classmethod
    def get_required_parameters(
        cls, input_streaming: bool
    ) -> Dict[str, Union[ValidInputTypes, Type[Iterable[ValidInputTypes]]]]:
        """Get the set of input types required by this task"""
        if not input_streaming:
            if _UNARY_PARAMS_ANNOTATION not in cls.__annotations__:
                raise ValueError("No unary inputs are specified for this task")
            return cls.__annotations__[_UNARY_PARAMS_ANNOTATION]
        if _STREAM_PARAMS_ANNOTATION not in cls.__annotations__:
            raise ValueError("No streaming inputs are specified for this task")
        return cls.__annotations__[_STREAM_PARAMS_ANNOTATION]

    @classmethod
    def get_output_type(cls, output_streaming: bool) -> Type[DataBase]:
        """Get the output type for this task

        NOTE: This method is automatically configured by the @task decorator
            and should not be overwritten by child classes.
        """
        if not output_streaming:
            if _UNARY_OUT_ANNOTATION not in cls.__annotations__:
                raise ValueError("No unary outputs are specified for this task")
            return cls.__annotations__[_UNARY_OUT_ANNOTATION]
        if _STREAM_OUT_ANNOTATION not in cls.__annotations__:
            raise ValueError("No streaming outputs are specified for this task")
        return cls.__annotations__[_STREAM_OUT_ANNOTATION]

    @classmethod
    def _raise_on_wrong_output_type(cls, output_type, module, output_streaming: bool):
        task_output_type = cls.get_output_type(output_streaming)

        if cls._subclass_check(output_type, task_output_type):
            # Basic case, same type or subclass of it
            return

        if typing.get_origin(output_type) == Union:
            for union_type in typing.get_args(output_type):
                if cls._subclass_check(union_type, task_output_type):
                    # Something in the union has an acceptable type
                    return

        # Do some streaming checks
        if output_streaming:
            if cls._is_iterable_type(output_type):
                # task_output_type is already guaranteed to be Iterable[T]
                streaming_type = typing.get_args(task_output_type)[0]

                for iterable_type in typing.get_args(output_type):
                    if cls._subclass_check(iterable_type, streaming_type):
                        return

        raise TypeError(
            f"Wrong output type for module {module}: "
            f"Found {output_type} but expected {task_output_type}"
        )

    @staticmethod
    def _subclass_check(this_type, that_type):
        """Wrapper around issubclass that first checks if both args are classes.
        Returns True if the types are the same, or they are both classes and this_type
        is a subclass of that_type
        """
        if this_type == that_type:
            return True
        if isclass(this_type) and isclass(that_type):
            return issubclass(this_type, that_type)
        return False

    @staticmethod
    def _is_iterable_type(typ: Type) -> bool:
        """Returns True if typ is an iterable type.
        Does not work for types like `list`, `tuple`, but we're interested here in `List[T]` etc.

        This is implemented this way to support older python versions where
        isinstance(typ, typing.Iterable) does not work
        """
        try:
            iter(typ)
            return True
        except TypeError:
            return False


def task(
    unary_parameters: Dict[str, ValidInputTypes] = None,
    streaming_parameters: Dict[str, Type[Iterable[ValidInputTypes]]] = None,
    unary_output_type: Type[DataBase] = None,
    streaming_output_type: Type[Iterable[Type[DataBase]]] = None,
    **kwargs,
) -> Callable[[Type[TaskBase]], Type[TaskBase]]:
    """The decorator for AI Task classes.

    This defines an output data model type for the task, and a minimal set of required inputs
    that all public models implementing this task must accept.

    As an example, the `caikit.interfaces.nlp.SentimentTask` might look like::

        @task(
            unary_parameters={
                "raw_document": caikit.interfaces.nlp.RawDocument
            },
            streaming_parameters={
                "raw_documents": Iterable[caikit.interfaces.nlp.RawDocument]
            }
            unary_output_type=caikit.interfaces.nlp.SentimentPrediction
            streaming_output_type=Iterable[caikit.interfaces.nlp.SentimentPrediction]
        )
        class SentimentTask(caikit.TaskBase):
            pass

    and a module that implements this task might have methods like::

        @module(id="b9d98408-84c2-488c-8385-9d698effe60b", task=SentimentTask)
        class MyModule(ModuleBase):

            @SentimentTask.taskmethod()
            def run(raw_document: caikit.interfaces.nlp.RawDocument,
                    inference_mode: str = "fast") ->
                        caikit.interfaces.nlp.SentimentPrediction:
                # impl

            @SentimentTask.taskmethod(input_streaming=True, output_streaming=True)
            def run_bidi_stream(raw_documents: DataStream[caikit.interfaces.nlp.RawDocument])
                    -> DataStream[caikit.interfaces.nlp.SentimentPrediction]:
                # impl

    Note the run function may include other arguments beyond the minimal required inputs for
    the task.

    Args:
        unary_parameters (Dict[str, ValidInputTypes]): The required parameters that all module's
            unary-input inference methods must contain. A dictionary of parameter name to parameter
            type, where the types can be in the set of:
                - Python primitives
                - Caikit data models
                - Iterable containers of the above
                - Caikit model references (maybe?)
        streaming_parameters: The same as unary_parameters, but for streaming-input inference
            methods. All types must be in the form `Iterable[T]`

        unary_output_type (Type[DataBase]): The unary output type of the task, which all modules'
            unary-output inference methods must return. This must be a caikit data model type.
        streaming_output_type (Type[Iterable[Type[DataBase]]]): The streaming output type of the
            task, which all modules' streaming-output inference methods must return. This must be
            in the form Iterable[T].

    Returns:
        A decorator function for the task class, registering it with caikit's core registry of
            tasks.
    """

    def decorator(cls: Type[TaskBase]) -> Type[TaskBase]:
        error.subclass_check("<COR19436440E>", cls, TaskBase)

        # NB: python <= 3.9 safe way of setting class annotations
        cls_annotations = cls.__dict__.get("__annotations__", None)
        if cls_annotations is None:
            cls.__annotations__ = {}
            cls_annotations = cls.__dict__.get("__annotations__", None)

        if unary_parameters:
            cls_annotations[_UNARY_PARAMS_ANNOTATION] = unary_parameters
        if streaming_parameters:
            cls_annotations[_STREAM_PARAMS_ANNOTATION] = streaming_parameters
        if unary_output_type:
            cls_annotations[_UNARY_OUT_ANNOTATION] = unary_output_type
        if streaming_output_type:
            cls_annotations[_STREAM_OUT_ANNOTATION] = streaming_output_type

        # Backwards compatibility with old-style @tasks
        if "required_parameters" in kwargs and not unary_parameters:
            cls_annotations[_UNARY_PARAMS_ANNOTATION] = kwargs["required_parameters"]
        if "output_type" in kwargs and not unary_output_type:
            output_type = kwargs["output_type"]
            if cls._is_iterable_type(output_type):
                cls_annotations[_STREAM_OUT_ANNOTATION] = kwargs["output_type"]
            else:
                cls_annotations[_UNARY_OUT_ANNOTATION] = kwargs["output_type"]
        # End Backwards compatibility

        error.value_check(
            "<COR12671910E>",
            _UNARY_PARAMS_ANNOTATION in cls_annotations
            or _STREAM_PARAMS_ANNOTATION in cls_annotations,
            "At least one input type must be set on a task",
        )
        error.value_check(
            "<COR12671910E>",
            _UNARY_OUT_ANNOTATION in cls_annotations
            or _STREAM_OUT_ANNOTATION in cls_annotations,
            "At least one output type must be set on a task",
        )

        if _UNARY_OUT_ANNOTATION in cls_annotations:
            error.subclass_check(
                "<COR12766440E>", cls.get_output_type(output_streaming=False), DataBase
            )

        if _STREAM_OUT_ANNOTATION in cls_annotations:
            if typing.get_origin(cls.get_output_type(output_streaming=True)) is None:
                raise TypeError(
                    f"subclass check failed: {cls.get_output_type(output_streaming=True)} is \
                        not a subclass of (<class 'collections.abc.Iterable'>"
                )
            error.subclass_check(
                "<COR12766440E>",
                typing.get_origin(cls.get_output_type(output_streaming=True)),
                collections.abc.Iterable,
            )

        if _UNARY_PARAMS_ANNOTATION in cls_annotations:
            params_dict = cls.get_required_parameters(input_streaming=False)
            error.type_check("<COR19906440E>", dict, params_dict=params_dict)
            error.type_check_all(
                "<COR00123440E>", str, params_dict_keys=params_dict.keys()
            )
            # TODO: check proto-ability of things
        if _STREAM_PARAMS_ANNOTATION in cls_annotations:
            params_dict = cls.get_required_parameters(input_streaming=True)
            error.type_check("<COR19556230E>", dict, params_dict=params_dict)
            error.type_check_all(
                "<COR58796465E>", str, params_dict_keys=params_dict.keys()
            )
            for v in params_dict.values():
                error.subclass_check(
                    "<COR52740295E>", typing.get_origin(v), collections.abc.Iterable
                )

        return cls

    return decorator


def _make_keyname_for_module(module_class: Type) -> str:
    return ".".join([module_class.__module__, module_class.__qualname__])



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This package has classes that will serialize a python interface to a protocol buffer interface.

Typically used for `caikit.core.module`s that expose .train and .run functions.
"""
# Standard
from typing import Any, Dict, List, Optional, Type, Union, get_args, get_origin
import abc
import copy
import typing

# First Party
from py_to_proto.dataclass_to_proto import (  # NOTE: Imported from here for compatibility
    Annotated,
    FieldNumber,
)
import alog

# Local
from ...interfaces.common.data_model.stream_sources import S3Path
from . import protoable, type_helpers
from .compatibility_checker import ApiFieldNames
from .data_stream_source import make_data_stream_source
from .proto_package import snake_to_upper_camel
from caikit.core import ModuleBase, TaskBase
from caikit.core.data_model.base import DataBase
from caikit.core.data_model.dataobject import make_dataobject
from caikit.core.signature_parsing import CaikitMethodSignature, CustomSignature
from caikit.interfaces.runtime.data_model import ModelPointer, TrainingJob

log = alog.use_channel("RPC-SERIALIZERS")

INDENT = "    "


class CaikitRPCBase(abc.ABC):
    @property
    @abc.abstractmethod
    def module_list(self) -> List[Type[ModuleBase]]:
        """Return the list of caikit.core Modules that can be invoked using this RPC"""

    @property
    @abc.abstractmethod
    def request(self) -> "_RequestMessage":
        """Return the internal representation of the request message type for this RPC"""

    @property
    def name(self) -> str:
        return self._name

    def create_request_data_model(self, package_name: str) -> Type[DataBase]:
        """Dynamically create data model for this RPC's request message"""
        return self.request.create_data_model(package_name)

    def create_rpc_json(self, package_name: str) -> Dict:
        """Return json snippet for the service definition of this RPC"""
        rpc_json = {
            "name": f"{self.name}",
            "input_type": f"{package_name}.{self.request.name}",
            "output_type": self.return_type.get_proto_class().DESCRIPTOR.full_name,
        }
        return rpc_json


class ModuleClassTrainRPC(CaikitRPCBase):
    """Helper class to create a unique RPC corresponding with the train function
    for a given module class
    """

    def __init__(
        self,
        method_signature: CaikitMethodSignature,
    ):
        """Initialize a .proto generator with a single module to convert

        Args:
            method_signature (CaikitMethodSignature): The module method signature to
            generate an RPC for
        """
        self.clz: Type[ModuleBase] = method_signature.module
        self._method = ModuleClassTrainRPC._mutate_method_signature_for_training(
            method_signature
        )
        self._name = ModuleClassTrainRPC.module_class_to_rpc_name(self.clz)

        # Compute the mapping from argument name to type for the module's run
        log.debug3("Param Dict: %s", self._method.parameters)
        log.debug3("Return Type: %s", self._method.return_type)

        # Store the input and output protobuf message types for this RPC
        self.return_type = self._method.return_type

        # Inner / outer requests
        # The actual training parameters from the method signature are nested under `parameters`
        self._inner_request = _RequestMessage(
            ModuleClassTrainRPC.module_class_to_inner_request_name(self.clz),
            self._method.parameters,
            self._method.default_parameters,
        )

        params = {"model_name": str, "output_path": S3Path, "parameters": "PLACEHOLDER"}

        self._req = _RequestMessage(
            ModuleClassTrainRPC.module_class_to_req_name(self.clz),
            params,
            {},
        )

    def create_request_data_model(self, package_name: str):
        """Partial override of CaikitRPCBase.create_request_data_model to take care of the inner
        request data model"""
        # Build the inner request data model
        inner_request_data_model = self._inner_request.create_data_model(package_name)
        # Insert the new type into the outer request
        for triple_index, _ in enumerate(self._req.triples):
            if self._req.triples[triple_index][1] == "parameters":
                triple = self._req.triples[triple_index]
                self._req.triples[triple_index] = (
                    inner_request_data_model,
                    triple[1],
                    triple[2],
                )
        # Call the base class' method
        return CaikitRPCBase.create_request_data_model(
            self=self, package_name=package_name
        )

    @property
    def module_list(self) -> List[Type[ModuleBase]]:
        """Returns a list containing the single caikit.core.module type that this RPC is for"""
        return [self.clz]

    @property
    def request(self) -> "_RequestMessage":
        return self._req

    @staticmethod
    def module_class_to_rpc_name(module_class: Type[ModuleBase]) -> str:
        """Helper function to convert from the name of a module to the name of the
        request RPC function
        """
        return snake_to_upper_camel(
            f"{module_class.TASK_CLASS.__name__}_{module_class.__name__}_Train"
        )

    @staticmethod
    def module_class_to_req_name(module_class: Type[ModuleBase]) -> str:
        """Helper function to convert from the name of a module to the name of the
        request RPC message

        Example: self.clz._module__ = sample_lib.modules.sample_task.sample_implementation

        return: SampleTaskSampleModuleTrainRequest

        """
        return f"{ModuleClassTrainRPC.module_class_to_rpc_name(module_class)}Request"

    @staticmethod
    def module_class_to_inner_request_name(module_class: Type[ModuleBase]) -> str:
        """Helper function to convert from a module to the name of the inner message
        containing all the training parameters

        Example: self.clz._module__ = sample_lib.modules.sample_task.sample_implementation

        return: SampleTaskSampleModuleTrainParameters
        """
        return f"{ModuleClassTrainRPC.module_class_to_rpc_name(module_class)}Parameters"

    @staticmethod
    def _mutate_method_signature_for_training(
        signature: CaikitMethodSignature,
    ) -> Optional[CaikitMethodSignature]:
        # Change return type for async training interface
        return_type = TrainingJob

        # Loop through params and edit as needed
        new_params = {}
        for name, typ in signature.parameters.items():
            if type_helpers.has_data_stream(typ):
                # Assume this is training data
                # Save just the stream for now
                stream_type = type_helpers.get_data_stream_type(typ)
                element_types = get_args(stream_type)
                assert (
                    len(element_types) == 1
                ), "Cannot handle DataStream with multiple type args"
                element_type = element_types[0]
                new_params[name] = make_data_stream_source(element_type)
            elif type_helpers.is_model_type(typ):
                # Found a model pointer
                new_params[name] = ModelPointer
            else:
                new_params[name] = protoable.handle_protoables_in_union(
                    field_name=name,
                    arg_type=typ,
                )

        return CustomSignature(
            original_signature=signature, parameters=new_params, return_type=return_type
        )


class TaskPredictRPC(CaikitRPCBase):
    """Helper class to create a unique RPC for the aggregate set of Modules that
    implement the same task
    """

    def __init__(
        self,
        task: Type[TaskBase],
        method_signatures: List[CaikitMethodSignature],
        input_streaming: bool = False,
        output_streaming: bool = False,
    ):
        """Initialize a .proto generator with all modules of a given task to convert

        Args:
            task (Type[TaskBase]): Task type

            method_signatures (List[CaikitMethodSignature]): The list of method
                signatures from concrete modules implementing this task
        """
        self.task = task
        self._method_signatures = method_signatures
        self._input_streaming = input_streaming
        self._output_streaming = output_streaming

        # Aggregate the argument signature types into a single parameters_dict
        parameters_dict = {}
        default_parameters = {}
        for method in method_signatures:
            default_parameters.update(method.default_parameters)
            new_params = self._handle_task_inputs(method.parameters)
            primitive_arg_dict = protoable.to_protoable_signature(new_params)
            for arg_name, arg_type in primitive_arg_dict.items():
                current_val = parameters_dict.get(arg_name, arg_type)
                # TODO: raise runtime error here instead of assert!
                # TODO: need to resolve Optional[T] vs. T - if both come in, use Optional[T]
                assert (
                    current_val == arg_type
                ), f"Conflicting value types for arg {arg_name}: {current_val} != {arg_type}"

                parameters_dict[arg_name] = arg_type

        self._req = _RequestMessage(
            self._task_to_req_name(), parameters_dict, default_parameters
        )

        # Validate that the return_type of all modules in the grouping matches
        return_types = {
            protoable.get_protoable_return_type(method.return_type)
            for method in method_signatures
        }
        assert len(return_types) == 1, (
            f"Found multiple return types for task [{task}], rpc: [{self._task_to_rpc_name()}. "
            f"Return types: {return_types}]"
        )
        self.return_type = list(return_types)[0]

        # Create the rpc name based on the module type
        self._name = self._task_to_rpc_name()

    @property
    def module_list(self) -> List[Type[ModuleBase]]:
        """Returns the list of all caikit.core.modules that this RPC will be for. These should all
        be of the same ai-problem, e.g. my_caikit_library.modules.classification
        """
        return [method.module for method in self._method_signatures]

    @property
    def request(self) -> "_RequestMessage":
        return self._req

    @property
    def input_streaming(self) -> bool:
        return self._input_streaming

    @property
    def output_streaming(self) -> bool:
        return self._output_streaming

    def create_rpc_json(self, package_name: str) -> Dict:
        """Return json snippet for the service definition of this RPC"""
        if self.output_streaming:
            output_type_name = (
                typing.get_args(self.return_type)[0]
                .get_proto_class()
                .DESCRIPTOR.full_name
            )
        else:
            output_type_name = self.return_type.get_proto_class().DESCRIPTOR.full_name

        rpc_json = {
            "name": f"{self.name}",
            "input_type": f"{package_name}.{self.request.name}",
            "output_type": output_type_name,
            "server_streaming": self.output_streaming,
            "client_streaming": self.input_streaming,
        }
        return rpc_json

    def _handle_task_inputs(self, method_params: Dict[str, Any]) -> Dict[str, Any]:
        """Overrides input params with types specified in the Task"""
        new_params = {}
        if self._input_streaming:
            req_params = self.task.get_required_parameters(input_streaming=True)
            for param_name, param_type in method_params.items():
                if param_name in req_params:
                    # double check this condition, although it should already have been validated
                    # also assuming both are iterables
                    if get_args(req_params[param_name])[0] in get_args(param_type):
                        new_params[param_name] = get_args(req_params[param_name])[0]
                else:
                    new_params[param_name] = param_type
            return new_params
        # for unary input cases
        req_params = self.task.get_required_parameters(input_streaming=False)
        for param_name, param_type in method_params.items():
            if param_name in req_params:
                new_params[param_name] = req_params[param_name]
            else:
                new_params[param_name] = param_type
        return new_params

    def _task_to_req_name(self) -> str:
        """Helper function to convert the pair of library name and task name to
        a request message name
        """
        if self._input_streaming and self._output_streaming:
            return snake_to_upper_camel(f"BidiStreaming{self.task.__name__}_Request")
        if self._output_streaming:
            return snake_to_upper_camel(f"ServerStreaming{self.task.__name__}_Request")
        if self._input_streaming:
            return snake_to_upper_camel(f"ClientStreaming{self.task.__name__}_Request")
        return snake_to_upper_camel(f"{self.task.__name__}_Request")

    def _task_to_rpc_name(self) -> str:
        """Helper function to convert the pair of library name and task name
        to an RPC name

        Example: self.task = (sample_lib, sample_task)

        return: SampleTaskPredict
        """
        if self._input_streaming and self._output_streaming:
            return snake_to_upper_camel(f"BidiStreaming{self.task.__name__}_Predict")
        if self._output_streaming:
            return snake_to_upper_camel(f"ServerStreaming{self.task.__name__}_Predict")
        if self._input_streaming:
            return snake_to_upper_camel(f"ClientStreaming{self.task.__name__}_Predict")
        return snake_to_upper_camel(f"{self.task.__name__}_Predict")


class _RequestMessage:
    """Helper class to create the input request message that wraps up the inputs
    for a given function. The request Contains N named data-model or primitive
    objects.
    """

    def __init__(
        self, msg_name: str, params: Dict[str, Type], default_map: Dict[str, Any]
    ):
        """Initialize with the module class and the parsed parameters to the run
        function.
        """
        self.name = msg_name
        self.triples = []
        self.default_map = default_map

        existing_fields = ApiFieldNames.get_fields_for_message(self.name)

        if len(existing_fields) > 0:
            last_used_number = max(existing_fields.values())
        else:
            last_used_number = 0

        for _, (item_name, typ) in enumerate(params.items()):
            if item_name in existing_fields:
                # if field existed previously, get the original number from there
                num = existing_fields[item_name]
            else:
                num = last_used_number + 1
                if get_origin(typ) is Union:
                    last_used_number += len(get_args(typ))
                else:
                    last_used_number += 1
            self.triples.append((typ, item_name, num))

        self.triples.sort(key=lambda x: x[2])

    def create_data_model(self, package_name: str) -> Type[DataBase]:
        """Dynamically create a data model for this request message"""
        properties = {
            # triple e.g. ('caikit.interfaces.common.ProducerPriority', 'producer_id', 1)
            # This does not take care of nested descriptors
            triple[1]: Annotated[triple[0], FieldNumber(triple[2])]
            for triple in self.triples
            if triple[1] not in self.default_map
        }
        optional_properties = {
            triple[1]: Annotated[Optional[triple[0]], FieldNumber(triple[2])]
            for triple in self.triples
            if triple[1] in self.default_map
        }
        attrs = copy.copy(self.default_map)

        if not properties and not optional_properties:
            log.warning(
                "No arguments found for request %s. Cannot generate rpc",
                self.name,
            )
            return None

        return make_dataobject(
            package=package_name,
            name=self.name,
            attrs=attrs,
            annotations={**properties, **optional_properties},
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RQA has a build that does not get write permissions, so a whole bunch of things can fail

We want to make sure that at least `import caikit.core` does not barf, which it will if we aren't
careful
"""
# Standard
from unittest.mock import patch

with patch("os.makedirs") as makedirs:
    makedirs.side_effect = PermissionError("No files for you")

    # Make sure importing does not fail
    # pylint: disable=unused-import
    # Local
    import caikit.core



# Standard
from dataclasses import dataclass
from typing import List
import json
import os
import pickle
import shutil
import tempfile

# Third Party
import pytest

# Local
from caikit.core.data_model import DataObjectBase, dataobject
from caikit.core.data_model.streams.data_stream import DataStream
from caikit.interfaces.common.data_model.stream_sources import Directory, File, S3Files
from caikit.runtime.service_generation.data_stream_source import (
    DataStreamSourceBase,
    _make_data_stream_source_type_name,
    make_data_stream_source,
)
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from sample_lib.data_model.sample import SampleTrainingType
from tests.conftest import temp_config
import caikit

################################################
## Helper functions


def validate_data_stream(data_stream, length, data_item_type, data_item_length=None):
    assert isinstance(data_stream, DataStream)
    assert len(data_stream) == length
    assert sum(1 for _ in data_stream) == length

    for data_item in data_stream:
        assert isinstance(data_item, data_item_type)
        if data_item_length is not None:
            assert len(data_item) == data_item_length


################################################
# Normal Tests


def test_make_data_stream_source_type_name():
    assert "DataStreamSourceInt" == _make_data_stream_source_type_name(int)
    assert "DataStreamSourceFloat" == _make_data_stream_source_type_name(float)
    assert "DataStreamSourceStr" == _make_data_stream_source_type_name(str)
    assert "DataStreamSourceBool" == _make_data_stream_source_type_name(bool)
    assert "DataStreamSourceSampleTrainingType" == _make_data_stream_source_type_name(
        SampleTrainingType
    )


def test_multiple_make_data_stream_source():
    """Defining 2 data stream sources in any order should not fail"""

    # Create 2 data stream sources
    make_data_stream_source(int)
    make_data_stream_source(SampleTrainingType)

    stream_type = caikit.interfaces.common.data_model.DataStreamSourceInt
    ds = stream_type(jsondata=stream_type.JsonData(data=[1, 2, 3]))
    proto_repr = ds.to_proto()
    assert ds.from_proto(proto_repr).to_proto() == proto_repr
    assert stream_type.from_proto(proto_repr).to_proto() == proto_repr


def test_data_model_element_type(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    assert isinstance(stream_type._to_element_type({"number": 1}), SampleTrainingType)


def test_primitive_element_type(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceInt
    assert isinstance(stream_type._to_element_type(1), int)


def test_make_data_stream_source_types():
    assert issubclass(make_data_stream_source(int), DataStreamSourceBase)
    assert issubclass(make_data_stream_source(float), DataStreamSourceBase)
    assert issubclass(make_data_stream_source(str), DataStreamSourceBase)


def test_make_data_stream_source_empty(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type()
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 0, None)


def test_pickle_round_trip_primitive():
    """Make sure that a source wrapping a primitive type can be round-tripped
    cleanly with pickle
    """
    stream_source = make_data_stream_source(float)
    inst = stream_source(jsondata=stream_source.JsonData(data=[1.23, -2.34]))
    round_trip = pickle.loads(pickle.dumps(inst))
    assert round_trip.to_dict() == inst.to_dict()


def test_pickle_round_trip_data_model():
    """Make sure that a source wrapping a data model object can be round-tripped
    cleanly with pickle
    """

    @dataobject
    class Foo(DataObjectBase):
        foo: int
        bar: str

    stream_type = make_data_stream_source(Foo)
    inst = stream_type(
        jsondata=stream_type.JsonData(data=[Foo(1, "two"), Foo(2, "three")])
    )
    round_trip = pickle.loads(pickle.dumps(inst))
    assert round_trip.to_dict() == inst.to_dict()


def test_data_stream_source_as_data_stream():
    """Make sure that a DataStreamSource works exactly like a DataStream"""
    stream_source = make_data_stream_source(float)
    source_list = [1.23, -2.34]

    # Test inline jsondata
    inst = stream_source(jsondata=stream_source.JsonData(data=source_list))
    assert list(inst) == source_list

    # Test from file
    # NOTE: Plain JSON doesn't round trip because ijson returns a special
    #   Decimal wrapper around raw values
    with tempfile.NamedTemporaryFile(mode="w", suffix=".jsonl") as handle:
        handle.write("\n".join([str(val) for val in source_list]))
        handle.flush()
        inst = stream_source(file=stream_source.File(filename=handle.name))
        assert list(inst) == source_list


def test_data_stream_source_base_path():
    """Make sure that a globally configured base path is used"""
    stream_source = make_data_stream_source(int)
    source_data = [1, 2, 3, 4]
    with tempfile.TemporaryDirectory() as workdir:
        with temp_config({"data_streams": {"file_source_base": workdir}}):
            nested_dir = os.path.join("foo", "bar")
            full_nested_dir = os.path.join(workdir, nested_dir)
            os.makedirs(full_nested_dir)
            fname = os.path.join(nested_dir, "data.json")
            full_fname = os.path.join(workdir, fname)
            with open(full_fname, "w") as handle:
                handle.write(json.dumps(source_data))

            # Make sure it works with the relative file path
            assert (
                list(stream_source(file=stream_source.File(filename=fname)))
                == source_data
            )

            # Make sure it works with the absolute file path
            assert (
                list(stream_source(file=stream_source.File(filename=full_fname)))
                == source_data
            )

            # Make sure it works with the relative directory
            assert list(
                stream_source(
                    directory=stream_source.Directory(
                        dirname=nested_dir,
                        extension="json",
                    )
                )
            ) == [source_data]

            # Make sure it works with the absolute directory
            assert list(
                stream_source(
                    directory=stream_source.Directory(
                        dirname=full_nested_dir,
                        extension="json",
                    )
                )
            ) == [source_data]

            # Make sure bad paths still raise the right errors
            with pytest.raises(CaikitRuntimeException):
                list(
                    stream_source(file=stream_source.File(filename="invalid/path.json"))
                )
            with pytest.raises(CaikitRuntimeException):
                list(
                    stream_source(
                        directory=stream_source.Directory(dirname="invalid/path")
                    )
                )


#################
# JSON data
#################


def test_make_data_stream_source_jsondata(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(
        jsondata=stream_type.JsonData(
            data=[SampleTrainingType(1), SampleTrainingType(2)]
        )
    )
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 2, SampleTrainingType)


def test_make_data_stream_source_jsondata_other_task(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceInt
    ds = stream_type(jsondata=stream_type.JsonData(data=[1]))
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 1, int)


#################
# File
#################


def test_make_data_stream_source_jsonfile(sample_train_service, sample_json_file):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(file=File(filename=sample_json_file))
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 2, SampleTrainingType)


def test_make_data_stream_source_csvfile(sample_train_service, sample_csv_file):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(file=File(filename=sample_csv_file))
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 3, SampleTrainingType)


def test_make_data_stream_source_jsonlfile(sample_train_service, sample_jsonl_file):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(file=File(filename=sample_jsonl_file))
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 4, SampleTrainingType)


def test_make_data_stream_source_from_file_with_no_extension(
    sample_train_service, sample_json_file, sample_jsonl_file, sample_csv_file, tmp_path
):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    for file in [sample_json_file, sample_jsonl_file, sample_csv_file]:
        no_extension_filename = os.path.join(str(tmp_path), "no_extension")
        shutil.copyfile(file, no_extension_filename)

        ds = stream_type(file=File(filename=no_extension_filename))
        assert isinstance(ds, DataStreamSourceBase)

        data_stream = ds.to_data_stream()
        assert isinstance(data_stream, DataStream)

        assert isinstance(data_stream.peek(), SampleTrainingType)


def test_make_data_stream_source_from_multipart_formdata_file(
    sample_train_service,
    sample_multipart_json,
    sample_multipart_csv,
    sample_multipart_json_with_content_header,
    tmp_path,
):
    """Test multipart streams. NB: We expect that multipart files will not have an extension"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType

    for file in (
        sample_multipart_csv,
        sample_multipart_json,
        sample_multipart_json_with_content_header,
    ):
        # prepare file with no extension
        no_extension_filename = os.path.join(str(tmp_path), "no_extension")
        shutil.copyfile(file, no_extension_filename)

        ds = stream_type(file=File(filename=no_extension_filename))
        assert isinstance(ds, DataStreamSourceBase)

        data_stream = ds.to_data_stream()
        assert isinstance(data_stream, DataStream)

        validate_data_stream(data_stream, 2, SampleTrainingType)


#################
# List of files
#################


def test_make_data_stream_source_list_of_json_files(
    sample_train_service, sample_json_file
):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    # does NOT work with sample_json_collection as each file needs to have data in array
    ds = stream_type(
        listoffiles=stream_type.ListOfFiles(files=[sample_json_file, sample_json_file])
    )
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 4, SampleTrainingType)


def test_make_data_stream_source_list_of_csv_files(
    sample_train_service, sample_csv_collection
):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    sample_files = [
        os.path.join(sample_csv_collection, file)
        for file in os.listdir(sample_csv_collection)
    ]
    ds = stream_type(listoffiles=stream_type.ListOfFiles(files=sample_files))
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 6, SampleTrainingType)


def test_make_data_stream_source_list_of_jsonl_files(
    sample_train_service, sample_jsonl_collection
):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    sample_files = [
        os.path.join(sample_jsonl_collection, file)
        for file in os.listdir(sample_jsonl_collection)
    ]
    ds = stream_type(listoffiles=stream_type.ListOfFiles(files=sample_files))
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 7, SampleTrainingType)


#################
# Directories
#################


def test_make_data_stream_source_jsondir(sample_train_service, sample_json_collection):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(directory=Directory(dirname=sample_json_collection))
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 3, SampleTrainingType)


def test_make_data_stream_source_csvdir(sample_train_service, sample_csv_collection):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(
        directory=Directory(dirname=sample_csv_collection, extension="csv")
    )
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 6, SampleTrainingType)


def test_make_data_stream_source_jsonldir(
    sample_train_service, sample_jsonl_collection
):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(
        directory=Directory(dirname=sample_jsonl_collection, extension="jsonl")
    )
    assert isinstance(ds, DataStreamSourceBase)

    data_stream = ds.to_data_stream()
    assert isinstance(data_stream, DataStream)

    validate_data_stream(data_stream, 7, SampleTrainingType)


def test_data_stream_operators():
    """Make sure that standard DataStream operators work as expected"""
    stream_type = make_data_stream_source(int)
    ds = stream_type(jsondata=stream_type.JsonData(data=[1, 2, 3, 4]))
    train, test = ds.train_test_split()
    assert sorted(list(train) + list(test)) == [1, 2, 3, 4]
    ds2 = DataStream.from_iterable([5, 6, 7])
    assert list(ds.chain(ds2)) == [1, 2, 3, 4, 5, 6, 7]
    assert list(ds.filter(lambda x: x % 2)) == [1, 3]
    assert len(list(ds.shuffle(2))) == 4
    assert list(ds.map(lambda x: x * 2)) == [2, 4, 6, 8]
    assert list(DataStream.from_iterable([ds]).flatten()) == [1, 2, 3, 4]
    zipped_ds = ds.zip(DataStream.from_iterable([5, 6, 7, 8]))
    assert list(zipped_ds) == [
        (1, 5),
        (2, 6),
        (3, 7),
        (4, 8),
    ]
    assert list(zipped_ds[0]) == [1, 2, 3, 4]


################################################
# Error Tests


def test_make_data_stream_source_invalid_file_raises(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(file=File(filename="invalid_file"))
    with pytest.raises(CaikitRuntimeException):
        ds.to_data_stream()


def test_make_data_stream_source_invalid_dir_raises(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(directory=Directory(dirname="invalid_dir"))
    with pytest.raises(CaikitRuntimeException):
        ds.to_data_stream()


def test_data_stream_source_single_oneof(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    with pytest.raises(CaikitRuntimeException):
        stream_type(
            jsondata=stream_type.JsonData(data=[SampleTrainingType(1)]),
            directory=Directory(dirname="invalid_dir"),
        )


def test_make_data_stream_source_invalid_ext_dir(sample_train_service):
    with tempfile.TemporaryDirectory() as tempdir:
        fname = os.path.join(tempdir, f"sample.txt")
        with open(fname, "w") as handle:
            handle.write("something")
        stream_type = (
            caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
        )
        ds = stream_type(directory=Directory(dirname=tempdir, extension="txt"))
        with pytest.raises(CaikitRuntimeException) as e:
            ds.to_data_stream()
        assert "Extension not supported!" in e.value.message


def test_make_data_stream_source_no_files_w_ext_dir(
    sample_train_service, sample_jsonl_collection
):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(
        directory=Directory(dirname=sample_jsonl_collection, extension="csv")
    )
    with pytest.raises(CaikitRuntimeException) as e:
        ds.to_data_stream()
    assert "contains no source files with extension" in e.value.message


def test_make_data_stream_source_non_json_array_errors(tmp_path):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    file_path = os.path.join(str(tmp_path), "non_array.json")
    with open(file_path, "w") as fp:
        fp.write(
            """
        {
            "number": 1,
            "label": "foo"
        }
        """
        )
    ds = stream_type(file=File(filename=file_path))
    assert isinstance(ds, DataStreamSourceBase)

    with pytest.raises(ValueError, match="Non-array JSON object"):
        ds.to_data_stream()


def test_s3_not_implemented(sample_train_service):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    ds = stream_type(s3files=S3Files())
    # Explicit .to_data_stream will fail
    with pytest.raises(
        NotImplementedError,
        match="S3Files are not implemented as stream sources in this runtime.",
    ) as e:
        ds.to_data_stream()

    # And so would iterating on the data stream source directly
    with pytest.raises(
        NotImplementedError,
        match="S3Files are not implemented as stream sources in this runtime.",
    ) as e:
        for val in ds:
            _ = val



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The LocalModelTrainer uses a local thread to launch and manage each training job
"""

# Standard
from concurrent.futures.thread import _threads_queues
from datetime import datetime, timedelta
from typing import Any, Dict, Iterable, List, Optional, Type, Union
import os
import re
import threading
import uuid

# First Party
import aconfig
import alog

# Local
from ...interfaces.common.data_model.stream_sources import S3Path
from ..data_model import TrainingStatus
from ..exceptions import error_handler
from ..modules import ModuleBase
from ..toolkit.destroyable_process import DestroyableProcess
from ..toolkit.destroyable_thread import DestroyableThread
from ..toolkit.logging import configure as configure_logging
from .model_trainer_base import ModelTrainerBase, TrainingInfo
from caikit.core.exceptions.caikit_core_exception import (
    CaikitCoreException,
    CaikitCoreStatusCode,
)
import caikit

log = alog.use_channel("LOC-TRNR")
error = error_handler.get(log)


# 🌶️🌶️🌶️
# Fix for python3.9, 3.10 and 3.11 issue where forked processes always exit with exitcode 1
# when it's created inside a ThreadPoolExecutor: https://github.com/python/cpython/issues/88110
# Fix taken from https://github.com/python/cpython/pull/101940
# Credit: marmarek, https://github.com/marmarek

if hasattr(os, "register_at_fork"):
    os.register_at_fork(after_in_child=_threads_queues.clear)


class LocalModelTrainer(ModelTrainerBase):
    __doc__ = __doc__

    name = "LOCAL"

    class LocalModelFuture(ModelTrainerBase.ModelFutureBase):
        """A local model future manages an execution thread for a single train
        operation
        """

        def __init__(
            self,
            trainer_name: str,
            module_class: Type[ModuleBase],
            save_path: Optional[Union[str, S3Path]],
            save_with_id: bool,
            model_name: Optional[str],
            external_training_id: Optional[str],
            use_subprocess: bool,
            subprocess_start_method: str,
            args: Iterable[Any],
            kwargs: Dict[str, Any],
        ):
            super().__init__(
                trainer_name=trainer_name,
                training_id=external_training_id or str(uuid.uuid4()),
                save_with_id=save_with_id,
                save_path=save_path,
                model_name=model_name,
                use_reversible_hash=external_training_id is None,
            )
            self._module_class = module_class

            # Placeholder for the time when the future completed
            self._completion_time = None
            # Set the submission time as right now. (Maybe this should be supplied instead?)
            self._submission_time = datetime.now()

            # Set up the worker and start it
            self._use_subprocess = use_subprocess
            self._subprocess_start_method = subprocess_start_method
            if self._use_subprocess:
                log.debug2("Running training %s as a SUBPROCESS", self.id)
                self._worker = DestroyableProcess(
                    start_method=self._subprocess_start_method,
                    target=self._train_and_save,
                    return_result=False,
                    args=args,
                    kwargs={
                        **kwargs,
                    },
                )
                # If training in a subprocess without a save path, the model
                # will be unreachable once trained!
                if not self.save_path:
                    log.warning(
                        "<COR28853922W>",
                        "Training %s launched in a subprocess with no save path",
                        self.id,
                    )
            else:
                log.debug2("Running training %s as a THREAD", self.id)
                self._worker = DestroyableThread(
                    self._train_and_save,
                    *args,
                    **kwargs,
                )
            self._worker.start()

        @property
        def completion_time(self) -> Optional[datetime]:
            return self._completion_time

        ## Interface ##

        def get_info(self) -> TrainingInfo:
            """Every model future must be able to poll the status of the
            training job
            """

            # The worker was canceled while doing work. It may still be in the
            # process of terminating and thus still alive.
            if self._worker.canceled:
                return self._make_training_info(status=TrainingStatus.CANCELED)

            # If the worker is currently alive it's doing work
            if self._worker.is_alive():
                return self._make_training_info(status=TrainingStatus.RUNNING)

            # The worker threw outside of a cancellation process
            if self._worker.threw:
                return self._make_training_info(
                    status=TrainingStatus.ERRORED, errors=[self._worker.error]
                )

            # The worker completed its work without being canceled or raising
            if self._worker.ran:
                return self._make_training_info(status=TrainingStatus.COMPLETED)

            # If it's not alive and not done, it hasn't started yet
            return self._make_training_info(status=TrainingStatus.QUEUED)

        def cancel(self):
            """Terminate the given training"""
            log.debug("Canceling training %s", self.id)
            with alog.ContextTimer(
                log.debug2, "Done canceling training %s in: ", self.id
            ):
                log.debug3("Destroying worker in %s", self.id)
                self._worker.destroy()

        def wait(self):
            """Block until the job reaches a terminal state"""
            log.debug2("Waiting for %s", self.id)
            self._worker.join()
            log.debug2("Done waiting for %s", self.id)
            self._completion_time = self._completion_time or datetime.now()

        def load(self) -> ModuleBase:
            """Wait for the training to complete, then return the resulting
            model or raise any errors that happened during training.
            """
            self.wait()
            if self._use_subprocess:
                log.debug2("Loading model saved in subprocess")
                error.value_check(
                    "<COR16745216E>",
                    self.save_path,
                    "Unable to load model from training {} "
                    + "trained in subprocess without a save_path",
                    self.id,
                )
                error.value_check(
                    "<COR59551640E>",
                    os.path.exists(self.save_path),
                    "Unable to load model from training {} "
                    + "saved in subprocess, path does not exist: {}",
                    self.id,
                    self.save_path,
                )
                result = caikit.load(self.save_path)
            else:
                result = self._worker.get_or_throw()
            return result

        ## Impl ##
        def _make_training_info(
            self, status: TrainingStatus, errors: Optional[List[Exception]] = None
        ) -> TrainingInfo:
            return TrainingInfo(
                status=status,
                errors=errors,
                completion_time=self._completion_time,
                submission_time=self._submission_time,
            )

        def _train_and_save(self, *args, **kwargs):
            """Function that will run in the worker thread"""
            # If running in a spawned subprocess, reconfigure logging
            if self._use_subprocess and self._subprocess_start_method != "fork":
                configure_logging()
            with alog.ContextTimer(log.debug, "Training %s finished in: ", self.id):
                trained_model = self._module_class.train(*args, **kwargs)
            if self.save_path is not None:
                log.debug("Saving training %s to %s", self.id, self.save_path)
                with alog.ContextTimer(log.debug, "Training %s saved in: ", self.id):
                    trained_model.save(self.save_path)
            self._completion_time = self._completion_time or datetime.now()
            log.debug2("Completion time for %s: %s", self.id, self._completion_time)
            return trained_model

    ## Interface ##

    # Expression for parsing retention policy
    _timedelta_expr = re.compile(
        r"^((?P<days>\d+?)d)?((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?((?P<seconds>\d*\.?\d*?)s)?$"
    )

    def __init__(self, config: aconfig.Config, instance_name: str):
        """Initialize with a shared dict of all trainings"""
        self._instance_name = instance_name
        self._use_subprocess = config.get("use_subprocess", False)
        self._subprocess_start_method = config.get("subprocess_start_method", "spawn")
        self._retention_duration = config.get("retention_duration")
        if self._retention_duration is not None:
            try:
                log.debug2("Parsing retention duration: %s", self._retention_duration)
                self._retention_duration = timedelta(
                    **{
                        key: float(val)
                        for key, val in self._timedelta_expr.match(
                            self._retention_duration
                        )
                        .groupdict()
                        .items()
                        if val is not None
                    }
                )
            except AttributeError:
                error(
                    "<COR63897671E>",
                    ValueError(
                        f"Invalid retention_duration: {self._retention_duration}"
                    ),
                )

        # The shared dict of futures and a lock to serialize mutations to it
        self._futures = {}
        self._futures_lock = threading.Lock()

    def train(
        self,
        module_class: Type[ModuleBase],
        *args,
        save_path: Optional[str] = None,
        save_with_id: bool = False,
        external_training_id: Optional[str] = None,
        model_name: Optional[str] = None,
        **kwargs,
    ) -> "LocalModelFuture":
        """Start training the given module and return a future to the trained
        model instance
        """
        # Always purge old futures
        self._purge_old_futures()

        # Wrap any models in the kwargs for safe spawning if needed
        if self._use_subprocess and self._subprocess_start_method != "fork":
            wrapped_models = {
                key: _SpawnProcessModelWrapper(val)
                for key, val in kwargs.items()
                if isinstance(val, ModuleBase)
            }
            log.debug2("Subprocess wrapped models: %s", wrapped_models.keys())
            kwargs.update(wrapped_models)

        # Create the new future
        model_future = self.LocalModelFuture(
            self._instance_name,
            module_class,
            save_path=save_path,
            save_with_id=save_with_id,
            external_training_id=external_training_id,
            use_subprocess=self._use_subprocess,
            subprocess_start_method=self._subprocess_start_method,
            model_name=model_name,
            args=args,
            kwargs=kwargs,
        )

        # Lock the global futures dict and add it to the dict
        with self._futures_lock:
            assert (
                model_future.id not in self._futures
            ), f"UUID collision for model future {model_future.id}"
            self._futures[model_future.id] = model_future

        # Return the future
        return model_future

    def get_model_future(self, training_id: str) -> "LocalModelFuture":
        """Look up the model future for the given id"""
        self._purge_old_futures()
        if model_future := self._futures.get(training_id):
            return model_future
        raise CaikitCoreException(
            status_code=CaikitCoreStatusCode.NOT_FOUND,
            message=f"Unknown training_id: {training_id}",
        )

    ## Impl ##

    def _purge_old_futures(self):
        """If a retention duration is configured, purge any futures that are
        older than the policy
        """
        if self._retention_duration is None:
            return
        now = datetime.now()
        purged_ids = {
            fid
            for fid, future in self._futures.items()
            if future.completion_time is not None
            and future.completion_time + self._retention_duration < now
        }
        if not purged_ids:
            log.debug3("No ids to purge")
            return
        log.debug3("Purging ids: %s", purged_ids)
        with self._futures_lock:
            for fid in purged_ids:
                # NOTE: Concurrent purges could have already done this, so don't
                #   error if the id is already gone
                self._futures.pop(fid, None)


class _SpawnProcessModelWrapper(ModuleBase):
    """This class wraps up a model to make it safe to pass to a spawned
    subprocess. It will not be efficient, but it will be safe!
    """

    def __init__(self, model: ModuleBase):
        super().__init__()
        self._model = model

    def __getattr__(self, name):
        """Forward attributes that are not found on the base class to the model

        NOTE: This does _not_ forward base class attributes since those are
            resolved before __getattr__ is called.
        """
        return getattr(self._model, name)

    def save(self, *args, **kwargs):
        """Directly forward save to the model so that it is not called by the
        base class
        """
        return self._model.save(*args, **kwargs)

    def run(self, *args, **kwargs):
        """Directly forward run to the model so that it is not called by the
        base class
        """
        return self._model.run(*args, **kwargs)

    def __getstate__(self) -> bytes:
        """When pickling, only send the serialized model body for non-fork. This
        is not a general-purpose pickle solution for models, but makes them safe
        for training jobs that need to move models between processes.
        """
        return self._model.as_bytes()

    def __setstate__(self, pickled: bytes):
        """When unpickling, deserialize the body if the model is not already
        loaded in the model manager. This must be used in conjunction with the
        above __getstate__ across a process boundary and should not be used as a
        general-purpose deserialization for models.
        """
        retrieved_model = caikit.core.load(pickled)
        self._model = retrieved_model



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from collections.abc import Iterable
import os
import tempfile

# Third Party
import pytest

# First Party
import alog

# Local
from caikit.core.exceptions import error_handler

# Unit Test Infrastructure
from tests.base import TestCaseBase
import caikit


class TestErrorHandler(TestCaseBase):
    def setUp(self):
        """Construct a new log channel and error handler."""
        self.log = alog.use_channel("TEST_ERR_HAN")
        self.error = error_handler.get(self.log)

    @pytest.fixture(autouse=True)
    def caplog(self, caplog):
        self.caplog = caplog

    def test_reuse_chan(self):
        """Verify that we are reusing the same channal when called with the same name."""
        for i in range(3):
            error = error_handler.get(self.log)
            self.assertTrue(error is self.error)

    def test_call_directly(self):
        """Verify that calling an error handler is equivalent to calling the `.log_raises` method."""
        self.assertTrue(
            error_handler.ErrorHandler.log_raise is error_handler.ErrorHandler.__call__
        )

    def test_log_raises_raises(self):
        """Verify that `log_raises` and `__call__` actually raises an exception."""
        with self.assertRaises(ValueError):
            self.error.log_raise("<XYZ>", ValueError("this is a test"))

        with self.assertRaises(FileNotFoundError):
            self.error("<YYY>", FileNotFoundError("this is a test"))

        with self.assertRaises(ValueError):
            self.error.log_raise(
                "<XYZ>", ValueError("this is a test"), RuntimeError("root error")
            )

        with self.assertRaises(ValueError):
            self.error(
                "<YYY>", ValueError("this is a test"), RuntimeError("root error")
            )

    def test_log_raises_logs(self):
        """Verify that `log_raises` actually logs an error message."""
        try:
            self.error("<AAA>", TypeError("this is a test"))
        except:
            pass

        # verify that one log line was omitted
        self.assertEqual(len(self.caplog.records), 1)

    def test_log_raises_max(self):
        """Verify that `log_raises` will not keep logging after max log lines written."""
        # raise the same exception over and over, max + 10 times
        exception = AttributeError("this is a test")
        for i in range(caikit.get_config().max_exception_log_messages + 10):
            try:
                self.error("<BBB>", exception)
            except:
                pass

        # verify that tracking attribute was set
        self.assertTrue(hasattr(exception, "_caikit_core_nexception_log_messages"))
        # verify that it was incremented for every raise
        self.assertEqual(
            exception._caikit_core_nexception_log_messages,
            caikit.get_config().max_exception_log_messages + 10 - 1,
        )
        # verify that it only wrote the max log lines
        # one for extra for message that logging will stop
        self.assertEqual(
            len(self.caplog.records), caikit.get_config().max_exception_log_messages + 1
        )

    def test_log_raises_max_with_root(self):
        """Verify that `log_raises` will not keep logging after max log lines written."""
        # raise the same exception over and over, max + 10 times
        exception = AttributeError("this is a test")
        root_exception = ValueError("this is the root ex")
        for i in range(caikit.get_config().max_exception_log_messages + 10):
            try:
                self.error("<BBB>", exception, root_exception)
            except:
                pass

        # verify that tracking attribute was set
        self.assertTrue(hasattr(exception, "_caikit_core_nexception_log_messages"))
        self.assertTrue(hasattr(root_exception, "_caikit_core_nexception_log_messages"))
        # verify that it was incremented for every raise
        self.assertEqual(
            exception._caikit_core_nexception_log_messages,
            caikit.get_config().max_exception_log_messages + 10 - 1,
        )
        self.assertEqual(
            root_exception._caikit_core_nexception_log_messages,
            caikit.get_config().max_exception_log_messages + 10 - 1,
        )

        # verify that it only wrote the max log lines
        # one for extra for message that logging will stop
        # Length of all records == 2 * MAX (one for each: exception and root exception)
        self.assertEqual(
            len(self.caplog.records),
            2 * (caikit.get_config().max_exception_log_messages + 1),
        )

    def test_type_check_raises(self):
        """Verify that `type_check` raises a `TypeError`."""
        with self.assertRaises(TypeError):
            self.error.type_check("<HHH>", int, float, bad_type="hello world")

    def test_type_check_passes(self):
        """Verify that `type_check` does not raise an exception when good."""
        try:
            self.error.type_check("<III>", int, float, good_type=42)
        except:
            self.assertTrue(False)

        try:
            self.error.type_check("<III>", int, float, good_type=42, another=33)
        except:
            self.assertTrue(False)

        try:
            self.error.type_check("<III>", int, str, good_type=42, another="abc")
        except:
            self.assertTrue(False)

    def test_type_check_none(self):
        """Verify that `type_check` does not raise when `None` values are allowed."""
        try:
            self.error.type_check(
                "<III>", str, allow_none=True, good_type=None, another="abc"
            )
        except:
            self.assertTrue(False)

    def test_type_check_logs(self):
        """Verify that `type_check` writes a single log line."""
        with self.assertRaises(TypeError):
            self.error.type_check("<JJJ>", int, bad_type="hello world")

        self.assertEqual(len(self.caplog.records), 1)

    def test_type_check_all_raises(self):
        """Verify that `type_check_all` raises a `TypeError`."""
        with self.assertRaises(TypeError):
            self.error.type_check_all("<ABC>", int, bad_type=(1, 2, "x"))

    def test_type_check_all_passes(self):
        """Verify that `type_check_all` does not reaise an exception when good."""
        try:
            self.error.type_check_all(
                "<III>", int, str, good_type=(42, "abc", "abc", 33)
            )
        except:
            self.assertTrue(False)

        try:
            self.error.type_check_all(
                "<III>",
                int,
                str,
                good_type=(
                    42,
                    "abc",
                ),
                another=[33, 33],
            )
        except:
            self.assertTrue(False)

    def test_type_check_all_iterable(self):
        """Verify that a custom Iterable passes type_check_all"""

        class CustomIterable(Iterable):
            def __init__(self, lst):
                self._lst = lst

            def __iter__(self):
                return self._lst.__iter__()

        self.error.type_check_all("<III>", int, good_type=CustomIterable([1, 2, 3]))

    def test_type_check_all_str(self):
        """Verify that a top level str object fails type_check_all"""
        with self.assertRaises(TypeError):
            self.error.type_check_all("<III>", str, str_type="<test_str>")

    def test_type_check_all_generator(self):
        """Verify that a top level generator object fails type_check_all"""
        with self.assertRaises(TypeError):
            self.error.type_check_all(
                "<III>", int, generator_type=(n for n in range(5))
            )

    def test_type_check_empty_passes(self):
        """Verify that `type_check_all` does not raise for empty variables."""
        try:
            self.error.type_check_all("<III>", int, str, good_type=(), another=[])
        except:
            self.assertTrue(False)

        try:
            self.error.type_check_all(
                "<III>", int, None, int, allow_none=True, good_type=(), another=[]
            )
        except:
            self.assertTrue(False)

        try:
            self.error.type_check_all(
                "<III>", int, good_type=(), another=[], again=[1, 2, 3]
            )
        except:
            self.assertTrue(False)

    def test_type_check_no_types(self):
        """Verify that a `RuntimeError` is raised if no types are passed."""
        with self.assertRaises(RuntimeError):
            self.error.type_check("<ABC>", foo="abc", bar=33)

    def test_type_check_no_variables(self):
        """Verify that a `RuntimeError` is raised if no variables are passed."""
        with self.assertRaises(RuntimeError):
            self.error.type_check("<ABC>", int, str, allow_none=True)

        with self.assertRaises(RuntimeError):
            self.error.type_check("<ABC>", int, str, allow_none=False)

    def test_type_check_all_non_list_or_tuple(self):
        """Verify that `type_check_all` raises a `TypeError` if variable is not a list or tuple.."""
        with self.assertRaises(TypeError):
            self.error.type_check_all("<ABC>", int, float, foo="abc")

        with self.assertRaises(TypeError):
            self.error.type_check_all("<ABC>", int, float, foo=3)

        with self.assertRaises(TypeError):
            self.error.type_check_all("<ABC>", int, float, foo={"a": 1, "b": 2})

    def test_type_check_all_logs(self):
        """Verify that `type_check_all` omitts a log line."""
        with self.assertRaises(TypeError):
            self.error.type_check_all("<ABC>", int, bad_type=(3.3, "abc"))

        self.assertEqual(len(self.caplog.records), 1)

    def test_type_check_all_none(self):
        """Verify that `type_check_all` allows `None` values, but not to contain `None` values."""
        try:
            self.error.type_check_all(
                "<XYZ>", int, allow_none=True, foo=(1, 2, 3), bar=None
            )
        except:
            self.assertTrue(False)

        with self.assertRaises(TypeError):
            self.error.type_check_all(
                "<XYZ>", int, allow_none=True, foo=(1, 2, None), bar=None
            )

    def test_type_check_all_no_types(self):
        """Verify that a `RuntimeError` is raised if no types are passed."""
        with self.assertRaises(RuntimeError):
            self.error.type_check_all("<ABC>", foo=(1, 2, 3), bar=("a", "b", "c"))

    def test_type_all_check_no_variables(self):
        """Verify that a `RuntimeError` is raised if no variables are passed."""
        with self.assertRaises(RuntimeError):
            self.error.type_check_all("<ABC>", int, str, allow_none=True)

        with self.assertRaises(RuntimeError):
            self.error.type_check_all("<ABC>", int, str, allow_none=False)

    def test_subclass_check_raises(self):
        """Verify that `subclass_check` raises a `TypeError`."""
        with pytest.raises(TypeError):
            self.error.subclass_check("<HHH>", int, float)

    def test_subclass_check_passes(self):
        """Verify that `subclass_check` does not raise an exception when good."""

        class Base:
            pass

        class Derived(Base):
            pass

        class SecondOrder(Derived):
            pass

        class Other:
            pass

        self.error.subclass_check("<III>", Derived, Base)
        self.error.subclass_check("<III>", SecondOrder, Base)
        self.error.subclass_check("<III>", Derived, Base, Other)
        self.error.subclass_check("<III>", None, Base, Other, allow_none=True)

    def test_subclass_check_none(self):
        """Verify that `subclass_check` does not raise when `None` values are allowed."""
        self.error.subclass_check("<III>", None, int, allow_none=True)

    def test_subclass_check_logs(self):
        """Verify that `subclass_check` writes a single log line."""
        with pytest.raises(TypeError):
            self.error.subclass_check("<JJJ>", int, float)

        self.assertEqual(len(self.caplog.records), 1)

    def test_subclass_requires_base_type(self):
        """Verify that `subclass_check` raises a RuntimeError if no base given"""
        with pytest.raises(RuntimeError):
            self.error.subclass_check("<JJJ>", int)

    def test_subclass_check_non_type(self):
        """Verify that `subclass_check` raises if the value is not a type"""
        with pytest.raises(TypeError):
            self.error.subclass_check("<III>", 1, int, allow_none=True)

    @pytest.mark.skipif(
        os.environ.get("LOG_FORMATTER") == "pretty", reason="Pretty formatter"
    )
    def test_value_check_raises(self):
        """Verify that `value_check` raises a `ValueError`."""
        bad_value = 1.1

        with self.assertRaises(ValueError):
            self.error.value_check(
                "<CCC>", 0.0 <= bad_value <= 1.0, "invalid range `{}`", bad_value
            )
        error_msg = self.caplog.records[0].msg["message"]
        self.assertEqual(
            error_msg,
            "exception raised: ValueError('value check failed: invalid range `1.1`')",
        )

    @pytest.mark.skipif(
        os.environ.get("LOG_FORMATTER") == "pretty", reason="Pretty formatter"
    )
    def test_value_check_no_args_raises(self):
        """Verify that `value_check` raises a `ValueError` if no args are given with empty msg."""
        bad_value = 1.1

        with self.assertRaises(ValueError):
            self.error.value_check("<CCC>", 0.0 <= bad_value <= 1.0)
        error_msg = self.caplog.records[0].msg["message"]
        self.assertEqual(
            error_msg,
            "exception raised: ValueError('value check failed: ')",
        )

    @pytest.mark.skipif(
        os.environ.get("LOG_FORMATTER") == "pretty", reason="Pretty formatter"
    )
    def test_value_check_empty_msg_raises(self):
        """Verify that `value_check` raises a `ValueError` if msg is empty."""
        bad_value = 1.1

        with self.assertRaises(ValueError):
            self.error.value_check("<CCC>", 0.0 <= bad_value <= 1.0, bad_value)
        error_msg = self.caplog.records[0].msg["message"]
        self.assertEqual(
            error_msg,
            "exception raised: ValueError('value check failed: 1.1')",
        )

    @pytest.mark.skipif(
        os.environ.get("LOG_FORMATTER") == "pretty", reason="Pretty formatter"
    )
    def test_value_check_empty_args_raises(self):
        """Verify that `value_check` raises a `ValueError` if only msg is provided."""
        bad_value = 1.1

        with self.assertRaises(ValueError):
            self.error.value_check("<CCC>", 0.0 <= bad_value <= 1.0, "invalid range")
        error_msg = self.caplog.records[0].msg["message"]
        self.assertEqual(
            error_msg,
            "exception raised: ValueError('value check failed: invalid range')",
        )

    def test_value_check_passes(self):
        """Verify that `value_check` does not raise an exception when good."""
        good_value = 0.5
        try:
            self.error.value_check(
                "<DDD>", 0.0 <= good_value <= 1.0, "invalid range `{}`", good_value
            )
        except:
            self.assertTrue(False)

    def test_value_check_logs(self):
        """Verify that `value_check` writes a single log line."""
        with self.assertRaises(ValueError):
            self.error.value_check("<EEE>", False, "invalid range")

        self.assertEqual(len(self.caplog.records), 1)

    def test_file_check_raises(self):
        """Verify that `file_check` fails when file is not found."""
        with self.assertRaises(FileNotFoundError):
            self.error.file_check(
                "<ABC>", "/etc/", "/hopefully/not/a/file/path", "/tmp"
            )

        with tempfile.TemporaryDirectory() as tempdir:
            with self.assertRaises(FileNotFoundError):
                self.error.file_check("<123>", tempdir)

    def test_file_check_logs(self):
        """Verify that `file_check` omitts a log message."""
        with self.assertRaises(FileNotFoundError):
            self.error.file_check("<XYZ>", "/definitely/not/a/valid/file/path")

        self.assertEqual(len(self.caplog.records), 1)

    def test_file_check_passes(self):
        """Verify that `file_check` does not raise when the file exists."""
        with tempfile.TemporaryDirectory() as tempdir:
            temp_filename = os.path.join(tempdir, "test.txt")

            with open(temp_filename, mode="w") as fh:
                fh.write("hello world")

            self.assertTrue(
                os.path.exists(temp_filename) and os.path.isfile(temp_filename)
            )

            try:
                self.error.file_check("<ABC>", temp_filename)
            except:
                self.assertTrue(False)

    def test_dir_check_raises(self):
        """Verify that `dir_check` fails when directory is not found."""
        with self.assertRaises(FileNotFoundError):
            self.error.dir_check("<ABC>", "/etc/", "/hopefully/not/a/dir/path", "/tmp")

    def test_dir_check_raises_on_file(self):
        """Verify that `dir_check` fails when pointed to a file."""
        with tempfile.TemporaryDirectory() as tempdir:
            temp_filename = os.path.join(tempdir, "test.txt")

            with open(temp_filename, mode="w") as fh:
                fh.write("hello world")

            self.assertTrue(
                os.path.exists(temp_filename) and os.path.isfile(temp_filename)
            )

            with self.assertRaises(NotADirectoryError):
                self.error.dir_check("<ABC>", temp_filename)

    def test_dir_check_logs(self):
        """Verify that `dir_check` omitts a log message."""
        with self.assertRaises(FileNotFoundError):
            self.error.dir_check("<XYZ>", "/definitely/not/a/valid/dir/path")

        self.assertEqual(len(self.caplog.records), 1)

    def test_dir_check_passes(self):
        """Verify that `dir_check` does not raise when the directory exists."""
        with tempfile.TemporaryDirectory() as tempdir:
            try:
                self.error.dir_check("<ABC>", tempdir)
            except:
                self.assertTrue(False)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from collections import Counter
from multiprocessing.pool import ThreadPool
from unittest.mock import patch
import json
import unittest

# Local
from caikit.runtime.metrics.rpc_meter import RPCMeter


class TestRPCMeter(unittest.TestCase):
    """This test suite tests the RPCMeter class which is using for RPC metering"""

    def setUp(self):
        """This method runs before each test begins to run"""
        self.rpc_meter = RPCMeter()
        # Note: These are not real classes in sample_lib
        self.another_widget_type = "<class 'sample_lib.modules.sample_task.sample_implementation.AnotherWidget'>"
        self.another_fidget_type = "<class 'sample_lib.modules.sample_task.sample_implementation.AnotherFidget'>"

    def test_update_metrics(self):
        pool = ThreadPool(1000)
        inputs = [self.another_widget_type for i in range(500)]
        inputs.extend([self.another_fidget_type for i in range(500)])
        pool.map(self.rpc_meter.update_metrics, inputs)

        self.assertEqual(len(self.rpc_meter.predict_rpc_counter), 1000)
        counter = Counter(self.rpc_meter.predict_rpc_counter)
        self.assertListEqual(
            list(counter.keys()), [self.another_widget_type, self.another_fidget_type]
        )
        self.assertListEqual(list(counter.values()), [500, 500])

    @patch("threading.Thread")
    def test_write_metrics(self, mock_thread):
        mock_thread.return_value = None
        with patch.object(
            self.rpc_meter, "predict_rpc_counter", [self.another_widget_type]
        ):
            self.rpc_meter._write_metrics()

        with patch.object(
            self.rpc_meter,
            "predict_rpc_counter",
            [self.another_fidget_type, self.another_fidget_type],
        ):
            self.rpc_meter._write_metrics()

        with patch.object(
            self.rpc_meter,
            "predict_rpc_counter",
            [self.another_widget_type, self.another_fidget_type],
        ):
            self.rpc_meter._write_metrics()

        data = []
        with open(self.rpc_meter.file_path) as f:
            data = [json.loads(line) for line in f]

        expected_keys = [
            "timestamp",
            "batch_size",
            "model_type_counters",
            "container_id",
        ]

        self.assertEqual(len(data), 3)
        self.assertListEqual(
            list(data[0].keys()),
            expected_keys,
        )
        self.assertEqual(data[0]["batch_size"], 1)
        self.assertEqual(len(data[0]["model_type_counters"]), 1)
        self.assertDictEqual(
            data[0]["model_type_counters"],
            {self.another_widget_type: 1},
        )

        self.assertEqual(data[1]["batch_size"], 2)
        self.assertEqual(len(data[1]["model_type_counters"]), 1)
        self.assertDictEqual(
            data[1]["model_type_counters"],
            {self.another_fidget_type: 2},
        )

        self.assertEqual(data[2]["batch_size"], 2)
        self.assertEqual(len(data[2]["model_type_counters"]), 2)
        self.assertDictEqual(
            data[2]["model_type_counters"],
            {self.another_widget_type: 1, self.another_fidget_type: 1},
        )

    def tearDown(self):
        self.rpc_meter.end_writer_thread()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file contains our logic about what constitutes a proto-able for RPC generation purposes
"""

# Standard
from typing import Dict, List, Type, Union, get_args, get_origin
import typing

# First Party
from py_to_proto.dataclass_to_proto import Annotated, OneofField
import alog

# Local
from caikit.core.data_model.base import DataBase
from caikit.core.data_model.dataobject import DATAOBJECT_PY_TO_PROTO_TYPES
from caikit.runtime.service_generation.type_helpers import is_data_model_type
import caikit

log = alog.use_channel("PROTOABLES")


def to_protoable_signature(signature: Dict[str, Type]) -> Dict[str, Type]:
    """Returns dictionary of protoable types only
    If there is a Union, pick the protoable type

    Args:
        signature (Dict[str, Type]): module signature of parameters and types
    """
    protoables = {}
    log.debug("Building protoable signature for %s", signature)
    for arg, arg_type in signature.items():
        protoable_type = handle_protoables_in_union(arg, arg_type)
        if protoable_type:
            protoables[arg] = protoable_type

    return protoables


def handle_protoables_in_union(field_name: str, arg_type: Type) -> Type:
    """Handles various protoable arg types from a Union.

    If arg_type is a union, then this will return the union back if all types in it are proto-able,
    or the first proto-able arg type if the union has non-protoable arg types.

    If arg_type is not a union nor protoable at all, this returns None.

    Examples:
    Union[protoable_type, non_protoable_type] -> protoable_type
    Union[protoable_type_1, protoable_type_2] -> Union[protoable_type_1, protoable_type_2]
    Union[protoable_type_1, protoable_type_2, non_protoable_type] -> protoable_type_1
    """
    if is_protoable_type(arg_type):
        if typing.get_origin(arg_type) == Union:
            union_protoables = [
                union_val
                for union_val in typing.get_args(arg_type)
                if is_protoable_type(union_val)
            ]
            # handle a union containing lists in a separate way
            if len(union_protoables) > 1 and any(
                typing.get_origin(arg) is list for arg in union_protoables
            ):
                return get_union_list_type(field_name, union_protoables)
            # if all are protoable, return the union (which will create a oneof)
            if len(union_protoables) == len(typing.get_args(arg_type)):
                return arg_type
            # if there's only 1 protoable found, return that
            if len(union_protoables) == 1:
                return union_protoables[0]
            # otherwise, try to get the data model objects in the Union
            dm_types = [arg for arg in union_protoables if is_data_model_type(arg)]
            # if there are multiple, pick the first one
            if len(dm_types) > 0:
                log.debug2(
                    "Picking first data model type %s in union protoables %s",
                    dm_types,
                    union_protoables,
                )
                return dm_types[0]
            log.debug(
                "Just picking first protoable type %s in union",
                union_protoables[0],
            )
            return union_protoables[0]
        return arg_type
    log.debug("Skipping non-protoable argument type [%s]", arg_type)


def get_union_list_type(field_name: str, union_protoables: List) -> Type[DataBase]:
    """Create a union from list type objects"""
    common_dm_package = caikit.interfaces.common.data_model
    param_list = []
    for arg in union_protoables:
        if get_origin(arg) is list:
            # Note: is_protoable_type ignores any list type without args
            arg_type = get_args(arg)[0]
            arg_name = f"{arg_type.__name__.capitalize()}Sequence"
            if not hasattr(common_dm_package, arg_name):
                raise AttributeError(
                    f"Unable to find {arg_name} in {common_dm_package}"
                )
            data_obj = getattr(common_dm_package, arg_name, None)
            if data_obj is None:
                raise AttributeError(
                    f"Unable to find {arg_name} in {common_dm_package}"
                )
            param_list.append(
                Annotated[
                    data_obj,
                    OneofField(field_name + "_" + arg_type.__name__ + "_" + "sequence"),
                ]
            )
        else:
            param_list.append(arg)
    return Union[tuple(param_list)]  # type: ignore


def get_protoable_return_type(arg_type: Type) -> Type:
    """Helper function that determines the right data model type to use from a Union"""

    # Decompose this type using typing to determine if it's a useful typing hint
    typing_origin = get_origin(arg_type)
    typing_args = get_args(arg_type)

    # If this is a data model type, no need to do anything
    if is_data_model_type(arg_type):
        return arg_type

    # Handle Unions by looking for a data model object in the union
    if typing_origin is Union:
        dm_types = [arg for arg in typing_args if is_data_model_type(arg)]
        if dm_types:
            log.debug2(
                "Found data model types in Union: [%s], taking first one", dm_types
            )
            return get_protoable_return_type(dm_types[0])

    # Handle iterables by returning `Iterable[T]`
    # py38 compatibility here
    try:
        iter(arg_type)
        if typing_origin:
            return typing.Iterable[typing_args]
    except TypeError:
        pass

    # if it's anything else we just return as is
    # we don't actually want to throw errors from service generation
    log.warning("Return type [%s] not a DM type, returning as is", arg_type)
    return arg_type


def is_protoable_type(arg_type: Type) -> bool:
    """
    Returns True if arg_type is in PROTO_TYPE_MAP(float, int, bool, str, bytes)
    Or if it's an imported Caikit data model class.
    Or if it's a Union of at least one of those.
    Or if it's a List of one of those.
    Or if it's a Dict of one of those.
    False otherwise"""
    proto_primitive_set = list(DATAOBJECT_PY_TO_PROTO_TYPES.keys())
    protoable = False
    if arg_type in proto_primitive_set:
        protoable = True
    elif is_data_model_type(arg_type):
        protoable = True
    elif typing.get_origin(arg_type) == list:
        log.debug2("Arg is List")
        if len(typing.get_args(arg_type)) == 0:
            log.debug2("List annotation has no type")
            protoable = False
        else:
            protoable = typing.get_args(arg_type)[0] in proto_primitive_set
    elif typing.get_origin(arg_type) == dict:
        log.debug2("Arg is Dict")
        if len(typing.get_args(arg_type)) == 0:
            log.debug2("Dict annotation has no type")
            protoable = False
        else:
            protoable = (
                typing.get_args(arg_type)[0] in proto_primitive_set
                and typing.get_args(arg_type)[1] in proto_primitive_set
            )
    elif typing.get_origin(arg_type) == Union:
        log.debug2("Arg is Union")
        # pylint: disable=use-a-generator
        protoable = any([is_protoable_type(arg) for arg in typing.get_args(arg_type)])

    if not protoable:
        log.debug2("Arg is not protoable, arg_type: %s", arg_type)
    return protoable



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from typing import Union
import os

# First Party
import aconfig
import alog

# Local
from .. import toolkit
from ..exceptions import error_handler

log = alog.use_channel("MODULE_CFG")
error = error_handler.get(log)


class ModuleConfig(aconfig.Config):
    """Config object used by all modules for config loading, saving, etc."""

    # keys that are not allowed at the top-level module configuration (reserved for internal use)
    reserved_keys = ["model_path"]

    def __init__(self, config_dict):
        """Construct a new module configuration object from a dictionary of config options.

        Args:
            config_dict (dict): Dictionary or containing the module's
                configuration.

        Notes:
            The following keys are reserved and *must not* be specified at the top level of a
            configuration:

            model_path - reserved for storing the original location where the model was loaded from
        """
        super().__init__(config_dict, override_env_vars=False)

        # validate that reserved configuration items are not in the config_dict
        self_keys_lower = {key.lower() for key in self.keys()}
        for reserved_key in self.reserved_keys:
            if reserved_key.lower() in self_keys_lower:
                error(
                    "<COR80419305E>",
                    KeyError(
                        "Do not add `{}` as top-level key in `config.yml`. "
                        "This is for internal use only.".format(reserved_key)
                    ),
                )

        # 🌶️🌶️🌶️: Backwards compatibility for old-style `blocks`, `workflows`, and `resources`
        if not self.module_id:  # pylint: disable=access-member-before-definition
            log.warning(
                "No module_id found in config. Re-save the model to use module_id."
            )
            if self.block_id:
                log.debug("Detected legacy block_id in config")
                self.module_id = self.block_id
            elif self.workflow_id:
                log.debug("Detected legacy workflow_id in config")
                self.module_id = self.workflow_id
            elif self.resource_id:
                log.debug("Detected legacy resource_id in config")
                self.module_id = self.resource_id

        error.value_check(
            "<COR80418932E>",
            hasattr(self, "module_id"),
            "Invalid model_config, module_id must be specified.",
        )

    @classmethod
    def load(cls, model_path: Union[str, "ModuleConfig"]) -> "ModuleConfig":
        """Load a new module configuration from a directory on disk.

        Args:
            model_path (Union[str, ModuleConfig]): Path to model directory. At
                the top level of directory is `config.yml` which holds info
                about the model. Note that the model_path here is assumed to be
                operating system correct as a consequence of the way this method
                is invoked by the model manager.

        Returns:
            model_config (ModuleConfig): Instantiated ModuleConfig for model
                given model_path.
        """
        error.type_check("<COR71170339E>", str, cls, model_path=model_path)
        if isinstance(model_path, cls):
            return model_path

        # Validate config.yml
        config_path = os.path.join(model_path, "config.yml")
        if not (os.path.exists(config_path) and os.path.isfile(config_path)):
            # NOTE: Do not log this out with error handler, as we might try this function multiple
            # times in some special cases, e.g., when handling zip archives.
            raise FileNotFoundError(
                "Module path `{}` is not a directory with a `config.yml` file.".format(
                    model_path
                )
            )

        # Read the yaml to dict and construct a new config object
        config = cls(toolkit.load_yaml(config_path))

        # Error if model_path was in the config
        if config.model_path is not None:
            error(
                "<COR80166142E>",
                KeyError(
                    "Do not add `model_path` as top-level key in `config.yml`. "
                    "This is for internal use only."
                ),
            )

        # Mdd the model path to the config object
        config["model_path"] = model_path

        # Make sure module_id is found
        if config.module_id is None:
            error(
                "<COR82701436E>",
                KeyError(f"No module_id found in config at {model_path}"),
            )

        return config

    def save(self, model_path):
        """Save this module configuration to a top-level `config.yml` file in the specified
        model path.

        Args:  str
            Path to model directory.  The `config.yml` file will be written to this location.

        Notes:
            `model_path` must already exist!  This means you must create the directory outside of
            this routine.
        """
        # make operating-system correct
        model_path = os.path.normpath(model_path)

        # create the directory where this config will be saved
        os.makedirs(model_path, exist_ok=True)

        # full path to config.yml
        config_path = os.path.join(model_path, "config.yml")

        # cast self into a dict and make sure we have a copy
        config_dict = dict(self).copy()

        # remove any reserved keys, these will be reproduced at load time
        for reserved_key in self.reserved_keys:
            if reserved_key in config_dict:
                config_dict.pop(reserved_key)

        # write to file
        toolkit.save_yaml(config_dict, config_path)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import uuid


class CaikitRuntimeException(Exception):
    """Exception raised when something went wrong while running the Caikit Runtime.

    Args:
        status_code (grpc.StatusCode): gRPC status code for the exception
        message (str): relevant information regarding what went wrong
        metadata (dict): defaults to {}, but contains any other objects/values relevant to exception

    Attributes:
        Same as Args. See above.
    """

    def __init__(self, status_code, message, metadata=None):
        super().__init__(message)
        self.status_code = status_code
        self.message = message
        self.id = uuid.uuid4().hex
        # in order to prevent errors in get_error_log .update call
        if metadata and isinstance(metadata, dict):
            self.metadata = metadata
            self.metadata.update({"error_id": self.id})
        else:
            # metadata is of None type
            self.metadata = {"error_id": self.id}



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from caikit.interfaces.nlp.data_model import FinishReason, GeneratedTextResult

## Setup #########################################################################

dummy_generated_response = GeneratedTextResult(
    generated_text="foo bar", generated_tokens=2, finish_reason=FinishReason.TIME_LIMIT
)

## Tests ########################################################################

### Generated Text Result
def test_all_fields_accessible():
    generated_response = GeneratedTextResult(
        generated_text="foo bar",
        generated_tokens=2,
        finish_reason=FinishReason.STOP_SEQUENCE,
    )
    # assert all((hasattr(obj, field) for field in obj.fields))
    assert generated_response.generated_text == "foo bar"
    assert generated_response.generated_tokens == 2
    assert generated_response.finish_reason == FinishReason.STOP_SEQUENCE


def test_from_proto_and_back():
    new = GeneratedTextResult.from_proto(dummy_generated_response.to_proto())
    assert new.generated_text == "foo bar"
    assert new.generated_tokens == 2
    assert new.finish_reason == FinishReason.TIME_LIMIT.value


def test_from_json_and_back():
    new = GeneratedTextResult.from_json(dummy_generated_response.to_json())
    assert new.generated_text == "foo bar"
    assert new.generated_tokens == 2
    assert new.finish_reason == FinishReason.TIME_LIMIT.value



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from os import path
import sys

# First Party
import alog

# Local
from caikit.runtime.__main__ import main
import caikit

if __name__ == "__main__":
    models_directory = path.abspath(path.join(path.dirname(__file__), "models"))
    caikit.config.configure(
        config_dict={
            "merge_strategy": "merge",
            "runtime": {
                "local_models_dir": models_directory,
                "library": "text_sentiment",
                "grpc": {"enabled": True},
                "http": {"enabled": True},
            },
        }
    )

    sys.path.append(
        path.abspath(path.join(path.dirname(__file__), "../"))
    )  # Here we assume that `start_runtime` file is at the same level of the
    # `text_sentiment` package

    alog.configure(default_level="debug")

    main()
